"""
threedb.evaluators.classification
==================================

Implements a basic evaluator for classification-based tasks.
"""

import json
from typing import Dict, List, Tuple, Any

import torch as ch
from torch import nn, Tensor
from typeguard import check_type

from threedb.evaluators.base_evaluator import BaseEvaluator, Output, LabelType

class SimpleClassificationEvaluator(BaseEvaluator):
    """SimpleClassificationEvaluator

    A concrete implementation of the abstract
    :class:`threedb.evaluators.base_evaluator.BaseEvaluator` that is designed
    for classification tasks.
    """
    output_type = 'classes'
    KEYS = ['is_correct', 'loss', 'prediction']

    def __init__(self, *, topk: int, classmap_path: str):
        """Initialize an Evaluator for classification

        Parameters
        ----------
        topk : int

        classmap_path : str
            a path to a JSON file mapping model UIDs to class numbers.
        """
        super().__init__()
        self.crit: nn.Module = nn.CrossEntropyLoss()
        self.topk: int = topk
        self.uid_to_targets: Dict[str, LabelType] = json.load(open(classmap_path))
        check_type('uid_to_targets', self.uid_to_targets, Dict[str, LabelType])

    def get_segmentation_label(self, model_uid: str) -> int:
        label = self.uid_to_targets[model_uid]
        if isinstance(label, list):
            return label[0]
        return label

    def get_target(self, model_uid: str, render_output: Dict[str, Tensor]) -> LabelType:
        """See the docstring of
        :meth:`threedb.evaluators.base_evaluator.BaseEvaluator.get_target`.
        """
        return self.uid_to_targets[model_uid]

    def declare_outputs(self) -> Dict[str, Tuple[List[int], str]]:
        """See the docstring of
        :meth:`threedb.evaluators.base_evaluator.BaseEvaluator.declare_outputs`.

        Returns
        -------
        Dict[str, Tuple[List[int], type]]

        """
        return {
            'is_correct': ([], 'bool'),
            'loss': ([], 'float32'),
            'prediction': ([self.topk], 'int32')
        }

    def summary_stats(self, pred: ch.Tensor, label: LabelType) -> Dict[str, Output]:
        """Concrete implementation of
        :meth:`threedb.evaluators.base_evaluator.BaseEvaluator.summary_stats`
        (see that docstring for information on the abstract function). Returns
        correctness (binary value) and cross-entropy loss of the prediction.

        Parameters
        ----------
        pred : ch.Tensor
            The output of the inference model: expected to be a 1D tensor of
            size (n_classes).
        label : LabelType
            An integer or list of integers representing the target label

        Returns
        -------
        Dict[str, Output]
            A dictionary containing the results to log from this evaluator,
            namely the correctness and the cross-entropy loss.
        """
        if isinstance(label, int):
            label = [label]
        _, topk_inds = pred.topk(self.topk)
        is_correct = any([pred_lab in label for pred_lab in topk_inds])
        stacked_pred: Tensor = ch.stack([pred for _ in range(len(label))])
        im_loss: float = float(self.crit(stacked_pred, ch.tensor(label)).item())
        return {
            'is_correct': is_correct,
            'loss': im_loss,
            'prediction': topk_inds
        }

    def to_tensor(self, pred: Any, *_) -> Tensor:
        return pred

Evaluator = SimpleClassificationEvaluator



"""
threedb.controls.blender.occlusion
==================================

Cover the subject object with another object. An example config file using this control can be found here:
`<https://github.com/3db/3db/tree/main/examples/unit_tests/occlusion.yaml>`_. 
"""

from ...try_bpy import bpy
from math import tan
from typing import Any, Dict
import numpy as np
from pathlib import Path
from ..base_control import PreProcessControl
from . import utils

class OcclusionControl(PreProcessControl):
    """Control that adds an occlusion object infront of the
    main object in the scene.

    Continuous Dimensions:

    - ``occlusion_ratio``: Ratio of the occluded part of the object of interest.
      e.g., parameter 0.01 would occlude 1% of the object, and parameter 0.8 would occlude 80%
      of the object. Default range: ``(0.01, .8)``.
    - ``zoom``: Controls distance from the occluder to the object of interest. The occluder is
      placed between the object of interest and the camera at a distance of ``zoom * D``,
      measured from the object of interest. ``D`` is the distance between the object of interest
      and the camera. Default range: ``(0.01, 0.4)``.
    - ``scale``: rescale the occlusion object by a factor of ``scale``. Parameter 0.5 halves the
      occluder's size, and parameter 2 doubles the occluder's size. Default range: ``(0.01, 1)``.

    Discrete Dimensions:

    - ``direction``: The direction from which the occluder approaches the
      object of interest. Takes a value between 0 and 7
      represeting the indices of the ``DIRECTIONS`` vectors.
    - ``occluder``: The occlusion object. This is an index of the list of the
      occlusion objects which is automatically initialized when the
      OcclusionControl is created (see Note).

    .. admonition:: Example images

        .. thumbnail:: /_static/logs/occlusion/images/image_1.png
            :width: 100
            :group: occlusion

        .. thumbnail:: /_static/logs/occlusion/images/image_2.png
            :width: 100
            :group: occlusion

        .. thumbnail:: /_static/logs/occlusion/images/image_3.png
            :width: 100
            :group: occlusion

        .. thumbnail:: /_static/logs/occlusion/images/image_4.png
            :width: 100
            :group: occlusion

        .. thumbnail:: /_static/logs/occlusion/images/image_5.png
            :width: 100
            :group: occlusion

        Varying all the parameters.
    """
    DIRECTIONS = [( 1, -1), ( 1, 0), ( 1, 1),
                  ( 0, -1),          ( 0, 1),
                  (-1, -1), (-1, 0), (-1, 1)]
    OOD_DIR = 'ood_objects'

    def __init__(self, root_folder):
        """Initializes the `OcclusionControl`

        Parameters
        ----------
        root_folder
            The root folder where the `ood_objects` folder containing
            the possible occluders exist
        """
        continuous_dims = {
            "occlusion_ratio": (0.01, 0.8),
            "zoom": (0.01, 0.4),
            "scale": (0.01, 1),
        }

        ood_folder = Path(root_folder) / self.OOD_DIR
        occluders_paths = list(ood_folder.glob('*.blend'))
        occluders_paths = [f.name for f in occluders_paths]

        print('ood_folder', ood_folder)
        print('occluders_paths', occluders_paths)

        discrete_dims = {
            "direction": list(range(len(self.DIRECTIONS))),
            "occluder": occluders_paths
        }
        assert len(discrete_dims['occluder']) >= 1, 'No occluder objects found!'
        super().__init__(root_folder,
                         continuous_dims=continuous_dims,
                         discrete_dims=discrete_dims)
        self.occluder_paths = occluders_paths
        self.ood_folder = ood_folder

    def _move_in_plane(self, ob, x_shift, y_shift):
        """Shifts the `ob` object in a plane passing through the
        ob location, and parallel to the camera frame.

        Parameters
        ----------
        ob
            object of interest
        x_shift
            Shift along the X-axis in the camera frame.
            Takes a value between -1 and 1.
        Y_shift
            Shift along the Y-axis in the camera frame
            Takes a value between -1 and 1.
        """
        C = bpy.context

        aspect = C.scene.render.resolution_x / C.scene.render.resolution_y

        camera = C.scene.objects["Camera"]
        fov = camera.data.angle_y
        z_obj_wrt_camera = np.linalg.norm(camera.location - ob.location)
        y_limit = tan(fov / 2) * z_obj_wrt_camera
        x_limit = y_limit * aspect
        camera_matrix = np.array(C.scene.camera.matrix_world)
        shift = np.matmul(
            camera_matrix,
            np.array([[x_limit * x_shift, y_limit * y_shift, -z_obj_wrt_camera, 1]]).T,
        )
        ob.location = shift[:3]

    def _find_center(self, bb, bb_occ, dir, occlusion_ratio=0.1):
        """Finds the location of the occlusion object `bb_occ` so that
        it occludes a specific ratio of the object `bb` from a given
        direction

        Parameters
        ----------
        bb
            The occluded object
        bb_occ
            The occluder
        dir
            The direction of occlusion
        occlustion_raio
            The ratio of occlusion
        """
        bb_area = bb[2] * bb[3]
        bb_occ_area = bb_occ[2] * bb[3]
        overlap_area = min(bb_occ_area, bb_area * occlusion_ratio)

        bb_xc = bb[0] + bb[2] / 2
        bb_yc = bb[1] + bb[3] / 2

        aspect = bb[3] / bb[2]

        if dir in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
            penetration = (
                np.abs(dir[0]) * overlap_area / bb_occ[3]
                + np.abs(dir[1]) * overlap_area / bb_occ[2]
            )
            x_out = bb_xc + dir[0] * (bb_occ[2] / 2 + bb[2] / 2 - penetration)
            y_out = bb_yc + dir[1] * (bb_occ[3] / 2 + bb[3] / 2 - penetration)

        elif dir == (1, 1) or (-1, 1) or (1, -1) or (-1, -1):
            penetration_x = np.sqrt(overlap_area * aspect)
            penetration_y = overlap_area / penetration_x
            # assert penetration_x / penetration_y == aspect
            x_out = bb_xc + dir[0] * (bb_occ[2] / 2 + bb[2] / 2 - penetration_x)
            y_out = bb_yc + dir[1] * (bb_occ[3] / 2 + bb[3] / 2 - penetration_y)

        return (x_out, y_out)

    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        C = bpy.context
        no_err, msg = self.check_arguments(control_args)
        assert no_err, msg       

        ob = context["object"]
        occ_basename = control_args['occluder']
        model_path = self.ood_folder / occ_basename
        occluder = utils.load_model(model_path)
        self.occluder = C.scene.objects[occluder]
        self.occluder.location = ob.location + float(control_args['zoom']) * (C.scene.camera.location - ob.location)

        bb = utils.camera_view_bounds_2d(C.scene, C.scene.camera, ob)
        bb_occ = utils.camera_view_bounds_2d(C.scene, C.scene.camera, self.occluder)
        x_shift, y_shift = self._find_center(bb, bb_occ,
                                             self.DIRECTIONS[control_args['direction']],
                                             control_args['occlusion_ratio'])
        x_shift = (x_shift - C.scene.render.resolution_x / 2) / (
            C.scene.render.resolution_x / 2
        )
        y_shift = (y_shift - C.scene.render.resolution_y / 2) / (
            C.scene.render.resolution_y / 2
        )
        self._move_in_plane(self.occluder, x_shift, y_shift)
        self.occluder.scale = (control_args['scale'],) * 3

    def unapply(self, context: Dict[str, Any]) -> None:
        """Deletes the occlusion objects. This is important to avoid
        clutering the scene with unneeded objects at the subsequent frames.

        Parameters
        ----------
        context : Dict[str, Any]
            The scene context object
        """
        bpy.ops.object.delete({"selected_objects": [self.occluder]})

Control = OcclusionControl



"""
threedb.evaluators.base_evaluator
==================================

Provides an abstract base class for implementing evaluators.

See :class:`threedb.evaluators.classification.SimpleClassificationEvaluator` for further details.
"""
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple, Union, Any

from torch import Tensor

Output = Union[bool, int, float, str]
LabelType = Union[List[int], int]

class BaseEvaluator(ABC):
    """Abstract Base Class for Evaluators

    An evaluator is in charge of taking the output of an inference model,
    analyzing it, and returning the relevant summary outputs (e.g.,
    correctness for classification, IoU for detection, MSE for segmentation,
    etc.)
    """
    KEYS: List[str] = ['is_correct', 'loss']

    @abstractmethod
    def get_segmentation_label(self, model_uid: str) -> int:
        """Given a model_uid, return a scalar label corresponding to the 3D
        model. This label is only used for the purpopses of generating a
        segmentation map. If the only goal is to separate the object from its
        background, this function can return anything greater than zero.

        Parameters
        ----------
        model_uid : str
            Unique identifier for the model.

        Returns
        -------
        int
            An integer which will be the color of the model in the segmentation
            map.
        """
        raise NotImplementedError

    @abstractmethod
    def get_target(self, model_uid: str, render_output: Dict[str, Tensor]) -> Any:
        """Given a model_uid, returns the corresponding label for that 3D model,
        which will be used to evaluate accuracy, loss, and other statistics.

        Parameters
        ----------
        model_uid : str
            The unique id of the model, as returned by (:meth:`threedb.rendering.base_renderer.BaseRenderer.render_and_apply`)
            (for blender, this will be the object ID)

        Returns
        -------
        int
            The class label for that object, most likely looked
            up through the evaluator metadata.
        """
        raise NotImplementedError

    @abstractmethod
    def summary_stats(self, pred: Any, label: Any) -> Dict[str, Output]:
        """Given a tensor corresponding to a model prediction on a single
        rendered image, and a tensor containing the label for that image, return
        a dictionary of summary statistics. The keys of the dictionary should
        match the ``KEYS`` static property of the concrete class.

        Parameters
        ----------
        pred : Any
            The output of the model being inspected with 3DB.
        label : Any
            The label corresponding to the rendered image.

        Returns
        -------
        Dict[str, Output]
            A dictionary mapping each key declared in ``type(self).KEYS`` to a
            serializable value (``int``, ``str``, ``float``, or ``bool``).
        """
        raise NotImplementedError

    @abstractmethod
    def declare_outputs(self) -> Dict[str, Tuple[List[int], str]]:
        """
        This function declares what the output of render() will be, based on the
        renderer settings. Returns a dictionary mapping keys to (dtype, size)
        tuples---the output of render() is string-to-tensor dictionary whose
        tensors will be checked against the return value of this function for
        both size and type.

        A basic implementation which suffices for most applications is provided
        in the abstract class :class:`threedb.rendering.base_renderer.BaseRenderer`.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_tensor(pred: Any, output_shape: List[int], input_shape: List[int]) -> Tensor:
        """Turns the output of the inference model into a PyTorch tensor. Useful
        for, e.g., detection models, where the model typically outputs a
        dictionary.

        Parameters
        ----------
        pred : Any
            The output of the inference model.
        output_shape : List[int]
            The desired shape of the output tensor
        input_shape : List[int]
            The shape of the image inputted into the inference model.

        Returns
        -------
        Tensor
            a Tensor representation of the model output
        """
        raise NotImplementedError



"""
threedb.controls.blender.background
===================================

Set the background to a solid-color. An example config file using this control can be found here:
`<https://github.com/3db/3db/tree/main/examples/unit_tests/background.yaml>`_.
"""

from typing import Any, Dict
from colorsys import hsv_to_rgb
import torch as ch

from ...try_bpy import bpy
from ..base_control import PostProcessControl

class BackgroundControl(PostProcessControl):
    """Control that replace the transparent background of a render (i.e., the
    alpha channel) with a given color specified in HSV by the control parameters.

    Continuous parameters:

    - ``H``, ``S`` and ``V``: the hue, saturation, and value of the color to
      fill the background with. (range: ``[0, 1]``)

    .. admonition:: Example images

        .. thumbnail:: /_static/logs/background/images/image_1.png
            :width: 100
            :group: background

        .. thumbnail:: /_static/logs/background/images/image_2.png
            :width: 100
            :group: background

        .. thumbnail:: /_static/logs/background/images/image_3.png
            :width: 100
            :group: background

        .. thumbnail:: /_static/logs/background/images/image_4.png
            :width: 100
            :group: background

        .. thumbnail:: /_static/logs/background/images/image_5.png
            :width: 100
            :group: background
        
        Varying all parameters across their ranges.
    """

    def __init__(self, root_folder: str):
        continuous_dims = {
            'H': (0., 1.),
            'S': (0., 1.),
            'V': (0., 1.),
        }
        super().__init__(root_folder,
                         continuous_dims=continuous_dims)

    def apply(self, render: ch.Tensor, control_args: Dict[str, Any]) -> ch.Tensor:
        check_result = self.check_arguments(control_args)
        assert check_result[0], check_result[1]
        bpy.context.scene.render.film_transparent = True

        rgb_color = hsv_to_rgb(control_args['H'], control_args['S'], control_args['V'])
        rgb_color = ch.tensor(rgb_color)[:, None, None].float()
        alpha = render[3:, :, :]
        img = render[:3, :, :] * alpha + (1 - alpha) * rgb_color
        return img

Control = BackgroundControl



"""
threedb.controls.blender.obj_loc_in_frame
=========================================

Change the object location in the frame. 
An example config file using this control can be found here:
`<https://github.com/3db/3db/tree/main/examples/unit_tests/obj_loc_in_frame.yaml>`_. 
"""

from ...try_bpy import bpy, mathutils
# from bpy import context as C
from math import tan
from typing import Any, Dict
import numpy as np
from ..base_control import BaseControl, PreProcessControl
from .utils import post_translate, cleanup_translate_containers

class ObjLocInFrameControl(PreProcessControl):
    """Control that moves the object of interest in the image frame.
    This is done by moving the object in a plane parallel to the camera
    plane and passing through the object's position.

    Continuous Dimensions:

    - ``x_shift``: The normalized X-coordinate of the center of the object in the frame.
        A value of -1 is the left-most edge of the frame and 1 is the right-most
        edge of the frame. (range: ``[-1, 1]``)
    - ``y_shift``: The normalized Y-coordinate of the center of the object in the frame
        Takes any value between -1 (bottom of the frame) and 1 (top of the frame).

    .. note::
        Setting x_shift and y_shift to zeros keeps the object in the
        middle of the frame.
    
    .. admonition:: Example images

        .. thumbnail:: /_static/logs/obj_loc_in_frame/images/image_1.png
            :width: 100
            :group: obj_loc_in_frame

        .. thumbnail:: /_static/logs/obj_loc_in_frame/images/image_2.png
            :width: 100
            :group: obj_loc_in_frame

        .. thumbnail:: /_static/logs/obj_loc_in_frame/images/image_3.png
            :width: 100
            :group: obj_loc_in_frame

        .. thumbnail:: /_static/logs/obj_loc_in_frame/images/image_4.png
            :width: 100
            :group: obj_loc_in_frame

        .. thumbnail:: /_static/logs/obj_loc_in_frame/images/image_5.png
            :width: 100
            :group: obj_loc_in_frame

        Example renderings while varying the parameters.
    
    """
    def __init__(self, root_folder: str):
        continuous_dims = {
            'x_shift': (-1., 1.),
            'y_shift': (-1., 1.),
        }
        super().__init__(root_folder, continuous_dims=continuous_dims)

    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        """Move the object in the frame

        Parameters
        ----------
        context
            The scene context.
        x_shift
            The normalized X-coordinate of the center of the object in the frame.
            Takes any value between -1 (left of the frame) and 1 (right of the frame).
        Y_shift
            The normalized Y-coordinate of the center of the object in the frame
            Takes any value between -1 (bottom of the frame) and 1 (top of the frame).

        """
        obj = context['object']
        bpy.context.view_layer.update()

        aspect = bpy.context.scene.render.resolution_x / bpy.context.scene.render.resolution_y
        camera = bpy.context.scene.objects['Camera']
        fov = camera.data.angle_y
        z_obj_wrt_camera = np.linalg.norm(camera.location - obj.location)

        y_limit = tan(fov/2) * z_obj_wrt_camera
        x_limit = y_limit * aspect

        camera_matrix = np.array(bpy.context.scene.camera.matrix_world)
        coords = [x_limit * control_args['x_shift'],
                  y_limit * control_args['y_shift'],
                  - z_obj_wrt_camera, 1]
        shift = np.matmul(camera_matrix, np.array([coords]).T)
        post_translate(obj, mathutils.Vector(list(shift[:3])))

    def unapply(self, context):
        cleanup_translate_containers(context['object'])

Control = ObjLocInFrameControl



"""
threedb.controls.blender.material
=================================

Change object materials. An example config file using this control can be found here:
`<https://github.com/3db/3db/tree/main/examples/unit_tests/material.yaml>`_. 
"""

from os import path
from glob import glob
from typing import Any, Dict
from pathlib import Path

from ...try_bpy import bpy
from ..base_control import PreProcessControl

MATERIAL_FOLDER = 'blender_control_material'

class MaterialControl(PreProcessControl):
    """Control that swap material of an object with another one

    Discrete Dimensions:

    - ``replacement_material``: The name of the material that replaces the
        original material.

    .. note::
        The possible values for ``replacement_material`` can be any file name
        found in ``$BLENDER_DATA/blender_control_material``.

        Each of these file should only contain a single material.

    .. admonition:: Example images

        .. thumbnail:: /_static/logs/material/images/image_1.png
            :width: 100
            :group: material

        .. thumbnail:: /_static/logs/material/images/image_2.png
            :width: 100
            :group: material

        .. thumbnail:: /_static/logs/material/images/image_3.png
            :width: 100
            :group: material

        .. thumbnail:: /_static/logs/material/images/image_4.png
            :width: 100
            :group: material

        .. thumbnail:: /_static/logs/material/images/image_5.png
            :width: 100
            :group: material
        
        Examples of various material substitutions.
    """
    def __init__(self, root_folder: str):
        mat_folder = Path(root_folder) / MATERIAL_FOLDER
        files_in_folder = mat_folder.glob('*.blend')
        files_in_folder = [f.name for f in files_in_folder]
        files_in_folder.append('keep_original')
        discrete_dims = {
            "replacement_material": files_in_folder
        }
        super().__init__(root_folder, discrete_dims=discrete_dims)


    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        """Replace the material of the target object with the material
        corresponding to ``replacement_material``.

        Parameters
        ----------
        context : Dict[str, Any]
            The blender scene context.
        control_args : Dict[str, Any]
            Must have key ``replacement_material`` containing the file name of the
            replacement material to use.
        """
        arg_check = self.check_arguments(control_args)
        assert arg_check[0], arg_check[1]

        state = {
            'added_materials': [],
            'replaced_materials': []
        }

        context['material_control_state'] = state

        replacement_material = control_args['replacement_material']
        if replacement_material == "keep_original":
            return

        fname = path.join(self.root_folder, MATERIAL_FOLDER, replacement_material)
        current_materials = set(x.name for x in bpy.data.materials)

        with bpy.data.libraries.load(fname) as (data_from, data_to):
            for material in data_from.materials:
                if material not in current_materials:
                    state['added_materials'].append(material)
                    data_to.materials.append(material)

        if len(state['added_materials']) != 1:
            raise Exception(f"File {fname} contains more than one material")

        replacing_with = state['added_materials'][0]

        for slot in context['object'].material_slots:
            try:
                state['replaced_materials'].append(slot.material.name)
            except Exception:
                state['replaced_materials'].append(None)

            slot.material = bpy.data.materials[replacing_with]


    def unapply(self, context: Dict[str, Any]) -> None:
        state = context['material_control_state']

        for slot, material_name in zip(context['object'].material_slots,
                                       state['replaced_materials']):

            if material_name is None:
                slot.material = None
            else:
                slot.material = bpy.data.materials[material_name]

        for material_name in state['added_materials']:
            bpy.data.materials.remove(bpy.data.materials[material_name])

Control = MaterialControl



"""
threedb.controls.blender.pointlight
===================================

EXPERIMENTAL CONTROL. DO NOT USE.
"""

from typing import Any, Dict
from ...try_bpy import bpy
import numpy as np
from colorsys import hsv_to_rgb
from threedb.controls.base_control import PreProcessControl

class PointLightControl(PreProcessControl):
    """This control adds a point light in the scene.

    Continuous Parameters:

    - ``H``, ``S``, ``V``: The color of the light (range: ``[0, 1]`` for each)
    - ``intensity``: The intensity of the light. Value depends on the environment.
      (range: ``[1000, 10000]``)
    - ``distance``: The distance away from the object of interest
    - ``dir_x``: relative (x, y, z)-coordinate of the point light w.r.t the object
      of interest. (ranges: ``[-1, 1]``, ``[-1, 1]``, ``[0, 1]`` respectively for x, y, z).

    .. note::
        1. You can add multiple point lights by using this control multiple
        times

        2. The vector ``(dir_x, dir_y, dir_z)`` is used along which direction
        w.r.t the is the point light placed, but the actual distance along
        that direction is decided by the ``distance`` parameter.

    """
    def __init__(self, root_folder: str):
        continuous_dims = {
            'H': (0, 1),
            'S': (0, 1),
            'V': (0, 1),
            'intensity': (1000, 10000),
            'distance': (5, 20),
            'dir_x': (-1, 1),
            'dir_y': (-1, 1),
            'dir_z': (0, 1),
        }
        super().__init__(root_folder, continuous_dims=continuous_dims)

    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        no_err, msg = self.check_arguments(control_args)
        assert no_err, msg

        bpy.ops.object.light_add(type='POINT', radius=1, align='WORLD')
        light_object = bpy.context.scene.objects["Point"]
        light = light_object.data
        obj = context['object']

        light_direction = np.array([control_args['dir_x'],
                                    control_args['dir_y'],
                                    control_args['dir_z']])
        light_direction = light_direction / np.linalg.norm(light_direction)

        # Set light location to point on sphere around object
        light_object.location = control_args['distance'] * light_direction + obj.location

        # Change light properties
        light.type = "POINT"
        light.color = hsv_to_rgb(control_args['H'],
                                 control_args['S'],
                                 control_args['V'])
        light.energy = control_args['intensity']

        # Optional - point light towards object
        bpy.ops.object.constraint_add(type="TRACK_TO")
        bpy.context.object.constraints["Track To"].target = obj

        return light_object

    def unapply(self, context: Dict[str, Any]) -> None:
        pass

Control = PointLightControl



"""
threedb.scheduling.base_scheduler
=================================

The task schedular of 3DB.
"""

from tqdm import tqdm
from typing import Any, Dict, Set, List
from threedb.scheduling.policy_controller import PolicyController
from threedb.scheduling.utils import recv_into_buffer
from threedb.result_logging.logger_manager import LoggerManager
from threedb.utils import CyclicBuffer

import time
import os
import json
import zmq

class Scheduler:
    def __init__(self, port: int,
                       max_running_policies: int,
                       envs: List[str],
                       models: List[str],
                       config: Dict[str, Dict[str, Any]],
                       policy_controllers: Set[PolicyController],
                       buffer: CyclicBuffer,
                       logger_manager: LoggerManager,
                       with_tqdm: bool = True) -> None:
        self.running = False

        self.envs = envs
        self.models = models
        self.buffer = buffer
        self.config = config
        self.logger_manager = logger_manager

        # Open a socket for communicating with the clients
        context = zmq.Context(io_threads=1)
        self.socket = context.socket(zmq.REP)
        self.socket.bind("tcp://*:%s" % port)

        # Keep track of the workers
        self.linked_workers: Set[str] = set()
        self.policy_controllers = policy_controllers
        self.num_policies = len(policy_controllers)
        self.done_policies = set()
        self.running_policies = set()
        self.max_running_policies = max_running_policies
        self.work_queue = {}

        # TQDM bars
        self.valid_renders, self.total_renders = 0, 0
        if with_tqdm:
            self.render_pb = tqdm(unit='images', desc='Renderings', smoothing=0.1)
            self.policies_pb = tqdm(unit='steps', desc='Policies', total=self.num_policies)

    def start(self, declared_outputs):
        """
        Parameters
        ----------
        - declared_outputs: map of key -> shape for what outputs the client will
          relay back to the server.
        
        Note
        ----
        Side effects: will create the buffer using the specified shapes, start
        the policy controllers, and also the loggers.
        """
        assert self.buffer.declare_buffers(declared_outputs)
        self.socket.send_pyobj({'kind': 'ack'})
        if not self.running:
            self.logger_manager.start()
            self.running = True

    def send_info(self):
        self.socket.send_pyobj({
            'kind': 'info',
            'environments': self.envs,
            'models': self.models,
            'render_args': self.config['render_args'],
            'inference': self.config['inference'],
            'controls_args': self.config['controls'],
            'evaluation_args': self.config['evaluation']
        })

    def handle_pull(self, message: Dict[str, Any]) -> None:
        """
        Handles a "pull" request from the client, asking for work. Should send a
        new list of jobs to work on
        """
        bs = message['batch_size']
        last_env = message['last_environment']
        last_model = message['last_model']

        to_work_on = []
        to_send = []

        def custom_order(arg):
            _, job, num_scheduled, time_scheduled = arg
            return (num_scheduled,
                    int(job.environment != last_env) + int(job.model != last_model),
                    time_scheduled, job.id)

        to_work_on = sorted(self.work_queue.values(), key=custom_order)[:bs]

        for _, job,  __, ___ in to_work_on:
            policy, job, num_scheduled, time_scheduled = self.work_queue[job.id]
            to_send.append(job)
            # Remember that we sent this job to one extra worker
            self.work_queue[job.id] = (policy, job, num_scheduled + 1, time_scheduled)

        # Send the job information to the worker node
        self.socket.send_pyobj({
            'kind': 'work',
            'params_to_render': to_send,
        })

    def handle_push(self, message: Dict[str, Any]) -> None:
        # Extract the result from the message
        jobid, result = message['job'], message['result']

        self.total_renders += 1

        if jobid in self.work_queue:
            # Recover the policy associated to this job entry
            selected_policy, job, _, _ = self.work_queue[jobid]
            del self.work_queue[job.id]  # This is done do not give it to anyone else 
            selected_policy.push_result(job.id, result)
            self.render_pb.update(1)
            self.valid_renders += 1
        else:
            # This task has been completed earlier by another worker
            self.buffer.free(result, -1) # We have to free the result

        self.socket.send_pyobj({'kind': 'ack'})

    def shutdown(self):
        for _ in tqdm(range(len(self.linked_workers)), desc='Shutting down', unit=' workers'):
            message = recv_into_buffer(self.socket, self.buffer)
            self.socket.send_pyobj({
                'kind': 'die'
            })
            try:
                self.linked_workers.remove(message['worker_id'])
            except:  # This worker didn't do work yet, we still shut it down
                pass

        self.buffer.close()
        self.render_pb.close()
        self.policies_pb.close()

        # Warning the logger that we are done
        self.logger_manager.log(None)
        print("==> [Waiting for any pending logging]")

        # We have to wait until it has processed everything left in the queue
        self.logger_manager.join()
        print("==> [Have a nice day!]")

    def schedule_work(self):
        wait_before_start_new = False

        while True:
            if len(self.done_policies) == self.num_policies:
                break  # We finished all the policies

            message = recv_into_buffer(self.socket, self.buffer)
            wid = message['worker_id']
            self.linked_workers.add(wid)

            if self.running:
                pulled_count = 0
                for policy in self.running_policies:
                    pulled = policy.pull_work()
                    if pulled is None: continue
                    pulled_count += 1
                    if pulled_count > 10:  break
                    wait_before_start_new = False
                    self.work_queue[pulled.id] = (policy, pulled, 0, time.time())

                # If there is not enough work we start a new policy
                little_work = len(self.work_queue) < 2 * len(self.linked_workers)
                policies_left = len(self.policy_controllers) > 0
                running_max_policies = len(self.running_policies) >= self.max_running_policies
                if little_work and (not wait_before_start_new) and policies_left and (not running_max_policies):
                    selected_policy = self.policy_controllers.pop()
                    selected_policy.start()
                    self.running_policies.add(selected_policy)
                    wait_before_start_new = True
            else:
                assert message['kind'] in {'info', 'decl'}, \
                    'message #1 was not "kind" == "info" or "decl", maybe race condition?'

            if message['kind'] == 'info':
                self.send_info()
            elif message['kind'] == 'decl':
                self.start(message['declared_outputs'])
            elif message['kind'] == 'pull':
                self.handle_pull(message)
            elif message['kind'] == 'push':
                self.handle_push(message)
            else:
                self.socket.send_pyobj({'kind': 'bad_query'})

            for policy in list(self.running_policies):
                if not policy.is_alive():
                    self.policies_pb.update(1)
                    self.running_policies.remove(policy)
                    self.done_policies.add(policy)

            self.render_pb.set_postfix({
                'workers': len(self.linked_workers),
                'pending': len(self.work_queue),
                'waste%': (1 - self.valid_renders / max(1e-10, self.total_renders)) * 100
            })
            self.policies_pb.set_postfix({'running': len(self.running_policies)})

        print("==> [Received all the results]")
        print("==> [Shutting down workers]")
        self.shutdown()



"""
threedb.result_logging.base_logger
==================================

Implements an abstract class for logging results.
"""

from multiprocessing import Process, Queue
from typing import Any, Dict, Optional
from threedb.utils import CyclicBuffer
from abc import abstractmethod, ABC

class BaseLogger(Process, ABC):
    """Abstract class for a Logger, inherits from
    ``multiprocessing.Process``. Implements the additional functions:

    - ``enqueue()``: put a new item on the logging queue, to be logged ASAP
    - ``log()``: the actual logging mechanism, meant to be overwritten by the
      user for each unique subclass. This should not be called directly,
      but rather will be called by ``run()``
    - ``run()``: main loop, waits for logs to be added to the queue, and calls
      ``log()`` on them.
    - ``end()``: Performs cleanup operations for the logger. No-op by default, should
        be overriden with code for closing any open file handles, ports, etc.

    """
    def __init__(self,
                 root_dir: str,
                 result_buffer: CyclicBuffer,
                 config: Optional[Dict[str, Dict[str, Any]]]) -> None:
        """Creates a logger instance

        Parameters
        ----------
        root_dir : str
            The directory in which to write the logging results (should be the
            same for all loggers, with each logger making a subfolder to log in)
        result_buffer : CyclicBuffer
            The buffer where the main thread is writing the results
        config : Optional[Dict[str, Dict[str, Any]]]
            The config file (parsed from YAML) of the 3DB experiment being run
        """        
        super().__init__()
        self.root_dir: str = root_dir
        self.buffer = result_buffer
        self.config: Optional[Dict[str, Dict[str, Any]]] = config
        self.queue: Queue = Queue()

    def enqueue(self, item: Dict[str, Any]) -> None:
        """Add an item in the queue to be logged. See `here`:TODO: for
        documentation on how the log item is structured.

        Parameters
        ----------
        item : Dict[str, Any]
            The item to be logged.
        """
        self.queue.put(item)

    @abstractmethod
    def log(self, item: Dict[str, Any]) -> None:
        """Abstract method for logging an item from the buffer

        Parameters
        ----------
        item : Dict[str, Any]
            A dictionary containing the results of a single rendering (as
            returned by :mod:`threedb.client`): see
            [TODO] for more detailed information on what this will contiain.
        """
        raise NotImplementedError

    @abstractmethod
    def end(self) -> None:
        """Performs cleanup operations for the logger. No-op by default, should
        be overriden with code for closing any open file handles, ports, etc.
        """
        pass

    def run(self) -> None:
        assert self.buffer.initialized, \
            'Tried to start logger before buffer was initialized'
        while True:
            item = self.queue.get()
            if item is None:
                break
            self.log(item)
        self.end()



"""
threedb.evaluators.detection
============================

An evaluator for object detection models.

Includes a concrete implementation of
:class:`threedb.evaluators.base_evaluator.BaseEvaluator`.
"""


import json
from typing import Any, Dict, List, Tuple

import torch as ch
from torch.tensor import Tensor
from torchvision.ops import boxes
from typeguard import check_type

from threedb.evaluators.base_evaluator import BaseEvaluator, LabelType, Output

# Default segmentation map value set by blender
BG_IND = -1

class SimpleDetectionEvaluator(BaseEvaluator):
    """Concrete implementation of
    :class:`threedb.evaluators.base_evaluator.BaseEvaluator`.
    """

    output_type = 'bboxes'
    output_shape = [100, 6]
    KEYS = ['is_correct_nolabel', 'precision_nolabel', 'recall_nolabel',
            'is_correct', 'precision', 'recall', 'is_valid']

    def __init__(self, iou_threshold: float, classmap_path: str,
                 min_recall: float = 1.0, min_precision: float = 0.0):
        super().__init__()
        self.iou_threshold = iou_threshold
        self.min_rec = min_recall
        self.min_prec = min_precision
        self.uid_to_targets: Dict[str, LabelType] = json.load(open(classmap_path))
        check_type('uid_to_targets', self.uid_to_targets, Dict[str, LabelType])

    def get_segmentation_label(self, model_uid: str) -> int:
        label = self.uid_to_targets[model_uid]
        return label[0] if isinstance(label, list) else label

    def declare_outputs(self) -> Dict[str, Tuple[List[int], str]]:
        return {
            'is_correct': ([], 'bool'),
            'precision': ([], 'float32'),
            'recall': ([], 'float32'),
            'is_correct_nolabel': ([], 'bool'),
            'precision_nolabel': ([], 'float32'),
            'recall_nolabel': ([], 'float32'),
            'is_valid': ([], 'bool')
        }

    def get_target(self,
                   model_uid: str,
                   render_output: Dict[str, Tensor]) -> ch.Tensor:
        """A label generator for object detection, using the outputted
        segmentation map. Uses a simplistic approach that assumes that each
        label in the segmentation map is a unique object. Specifically, for
        each unique label (object type) in the segmentation map, we make one
        bounding box, defined as the minimum-size box containing all pixels
        of this type in the segmentation map.

        Parameters
        ----------
        model_uid : str
            The UID of the model (not used)
        render_output : Dict[str, Tensor]
            The output of the renderer
            (:meth:`threedb.rendering.base_renderer.BaseRenderer.render_and_apply`)
            for this model. Must have a key called 'segmentation' containing
            the object segmentation.

        Returns
        -------
        ch.Tensor
            A set of bounding boxes and labels for the objects that should be
            detected. In particular, the bounding boxes are stored in ``(x1, y1,
            x2, y2, label)`` format.
        """
        seg_map = render_output['segmentation']
        unique_objs = list(map(int, seg_map.unique()))
        assert BG_IND in unique_objs, f'Background class ({BG_IND}) not found (found {unique_objs})'
        bbs = []
        for obj in unique_objs:
            if obj == BG_IND:
                continue
            filt = seg_map == obj
            _, cols, rows = ch.where(filt)
            bbs.append([rows.min(), cols.min(), rows.max(), cols.max(), obj])
        return ch.tensor(bbs)

    def summary_stats(self,
                      pred: Dict[str, ch.Tensor],
                      label: ch.Tensor) -> Dict[str, Output]:
        """Concrete implementation of
        :meth:`threedb.evaluators.base_evaluator.BaseEvaluator.summary_stats`

        Parameters
        ----------
        pred : Dict[str, ch.Tensor]
            Same output format as default torchvision detection models in
            evaluation mode, must have keys ('boxes', 'labels', 'scores')
        label : ch.Tensor
            Segmentation map containing the ground-truth objects

        Returns
        -------
        Dict[str, Output]
            The model's performance on the given image. Non-maximal suppression
            is performed on the output of the classifier, the precision and
            recall are calculated using the IOU threshold set at
            instantiation-time (see `here <https://en.wikipedia.org/wiki/Precision_and_recall>`_ 
            for information on precision and recall in object detection).

            Precision and recall are then thresholded (by the
            ``min_precision`` and ``min_recall`` parameters) to get a single
            boolean representing correctness. We return the corresponding
            keys ``precision_nolabel``, ``recall_nolabel``, and
            ``is_correct_nolabel`` as well as their counterparts
            ``precision``, ``recall`` and ``is_correct`` which take both box
            positions and class labels into account (the former only evaluate
            localization, not labelling).

            Finally, we return a key ``is_valid`` that represents whether we
            the label corresponding to the image is actually a valid class
            label.
        """
        keep_inds = boxes.nms(pred['boxes'], pred['scores'], self.iou_threshold)
        all_ious = boxes.box_iou(pred['boxes'][keep_inds], label[:, :4])
        iou_hits = (all_ious > self.iou_threshold)
        label_hits = label[:, 4][None].eq(pred['labels'][keep_inds][:, None])
        assert label_hits.shape == iou_hits.shape
        hits = iou_hits & label_hits
        rec_nl, prec_nl = [float(ch.any(iou_hits, dim=d).float().mean().item()) for d in (0, 1)]
        rec, prec = [float(ch.any(hits, dim=d).float().mean().item()) for d in (0, 1)]
        return {
            'is_correct_nolabel': (prec_nl >= self.min_prec) and (rec_nl >= self.min_rec),
            'precision_nolabel':  prec_nl,
            'recall_nolabel': rec_nl,
            'is_correct': (prec >= self.min_prec) and (rec >= self.min_rec),
            'precision': prec,
            'recall': rec,
            'is_valid': (int(label[:, 4]) != -1)
        }

    def to_tensor(self, pred: Any, output_shape: List[int], input_shape: List[int]) -> Tensor:
        """Concrete implementation of
        :meth:`threedb.evaluators.base_evaluator.BaseEvaluator.to_tensor`.

        Turns a prediction dictionary into a tensor with the given output_shape
        (N x 6). To do this, we concatenate the prediction into the form ``[(x1,
        y1, x2, y2, score, label)]``.
        """
        _, height, width = input_shape
        out = ch.zeros(*output_shape) - 1
        keep_inds = boxes.nms(pred['boxes'], pred['scores'], self.iou_threshold)
        num_boxes = keep_inds.shape[0]
        if num_boxes == 0:
            return out
        keys = ('boxes', 'scores', 'labels')
        kept_preds = [pred[s][keep_inds].view(num_boxes, -1).float() for s in keys]
        out[:num_boxes] = ch.cat(kept_preds, dim=1)
        out[:, [0, 2]] /= width
        out[:, [1, 3]] /= height
        return out

Evaluator = SimpleDetectionEvaluator



"""
threedb.controls.base_control
==============================

Abstract base classes for 3DB controls

The goal of control is two-fold:

- They declare a number of parameters, discrete or continuous.
- Based on values for these parameters they influence the image that will be
  rendered.

There are two kinds of Controls:

- *Pre-processing controls* update the scene before it is rendered (e.g., object
  orientation changes)
- *Post-processing controls* update the rendered 2D image (e.g., increasing
  contrast or adding noise)

This module file contains abstract base classes for both control types---new
controls can be introduced by sublcassing the appropriate one (see `here <custom_controls.html>`_ for
details).
"""

from abc import ABC, abstractmethod
from typing import Dict, Any, Optional, Tuple, List, Union

import torch as ch

class BaseControl(ABC):
    """An Abstract Base Class superclassing both pre- and post-processing
    controls. Implements the required cross-control properties
    ``continuous_dims`` and ``discrete_dims`` as well as a standard initializer.
    """

    @property
    def continuous_dims(self) -> Dict[str, Tuple[float, float]]:
        """Describes the set of continuous parameters this control needs.

        Returns
        -------
        Dict[str, Tuple[float, float]]
            Will be of the form: parameter_name -> (min_value, max_value)
        """
        return self._continuous_dims

    def update_continuous_dim(self, key: str, val: Tuple[float, float]):
        """Updates a specified continuous dimensions of the control with a
        user-provided value. Raises a ``ValueError`` if the key does not matched
        a pre-declared continuous control dimension.

        Parameters
        ----------
        key : str
            The key of the search dimension to override.
        val : Tuple[float, float]
            The new search space for that dimension.

        Raises
        ------
        ValueError
            If the key does not matched a pre-declared continuous dimension.
        """
        if not key in self._continuous_dims:
            valid_keys = self._continuous_dims.keys()
            raise ValueError(f'Unrecognized key {key} (expected one of {valid_keys})')
        self._continuous_dims[key] = val

    @property
    def discrete_dims(self) -> Dict[str, List[Any]]:
        """Describes the set of discrete parameters this control needs.

        Returns
        -------
        Dict[str, List[Any]]
            Will be of the form shape: parameter_name -> [value_1, value_2, ..., value_n]
        """
        return self._discrete_dims

    def update_discrete_dim(self, key: str, val: List[Any]):
        """Updates a specified discrete dimensions of the control with a
        user-provided value. Raises a ``ValueError`` if the key does not matched
        a pre-declared discrete control dimension.

        Parameters
        ----------
        key : str
            The key of the search dimension to override.
        val : List[Any]
            The new search space for that dimension.

        Raises
        ------
        ValueError
            If the key does not matched a pre-declared discrete dimension.
        """
        if not key in self._discrete_dims:
            valid_keys = self._discrete_dims.keys()
            raise ValueError(f'Unrecognized key {key} (expected one of {valid_keys})')
        self._discrete_dims[key] = val

    def __init__(self,
                 root_folder: str, *,
                 continuous_dims: Optional[Dict[str, Tuple[float, float]]] = None,
                 discrete_dims: Optional[Dict[str, List[Any]]] = None):
        """Construct a BaseControl

        Parameters
        ----------
        root_folder
            The folder containing all the data for this 3DB experiment. All
            paths are lative to his folder
        """
        self.root_folder = root_folder
        self._continuous_dims: Dict[str, Tuple[float, float]] = continuous_dims or {}
        self._discrete_dims: Dict[str, List[Any]] = discrete_dims or {}

    def check_arguments(self, control_args: Dict[str, Any]) -> Tuple[bool, str]:
        """Checks a dictionary of control arguments against the arguments
        pre-declared by ``self.continuous_dims`` and ``self.discrete_dims``

        Parameters
        ----------
        control_args : Dict[str, Any]
            A dictionary of control arguments.

        Returns
        -------
        Tuple[bool, str]
            A tuple indicating whether the control arguments are valid. If they
            are, the first return value is ``True`` and the second should be
            ignored. If not, the first return value is ``False`` and the second
            is an error message.
        """
        all_keys = self.continuous_dims.keys() | self.discrete_dims.keys()
        assert len(all_keys) == len(self.continuous_dims) + len(self.discrete_dims), \
            'Keys cannot be duplicated between continuous_dims and discrete_dims'
        if control_args.keys() != all_keys:
            return False, 'Keys in control arguments do not match declared keys'
        if not all(isinstance(control_args[k], float) or isinstance(control_args[k], int) for k in self.continuous_dims):
            return False, 'Some continuous arguments are not of type float'
        for k, v in self.discrete_dims.items():
            if not control_args[k] in v:
                return False, f'Argument {k} ({control_args[k]}) not in valid set: {v}'
        return True, ''

class PostProcessControl(BaseControl, ABC):
    """An abstract sublass of ``BaseControl`` for post-processing controls.
    Differs from pre-processing controls in that the type signature of
    ``apply``.
    """
    @abstractmethod
    def apply(self, render: ch.Tensor, control_args: Dict[str, Any]) -> ch.Tensor:
        """Modify a rendered image and return the transformed output.

        Parameters
        ----------
        render : ch.Tensor
            A tensor representation of the rendered image.
        control_args: Dict[str, Any]
            Control-specific settings (e.g., noise level for noise corruption,
            contrast level for contrast change, etc.).

        Returns
        -------
        ch.Tensor
            The post-processed output.
        """
        raise NotImplementedError

class PreProcessControl(BaseControl, ABC):
    """An abstract sublass of ``BaseControl`` for post-processing controls.
    Differs from pre-processing controls in that the type signature of
    ``apply``, and the presence of an ``unapply`` function undoing the effects
    of the control.
    """

    @abstractmethod
    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        """Modify the target give a combination of the inputs the control needs

        Parameters
        ----------
        context: Dict[str, Any]
            A dictionary representation of the current scene.
        control_args: Dict[str, Any]
            a dict containing an entry for each of the discrete and continuous
            parameters declared by the Control

        Modifies the scene in-place, no return value.
        """
        raise NotImplementedError

    @abstractmethod
    def unapply(self, context: Dict[str, Any]) -> None:
        """Undo the modification on a scene

        Note
        ----
        Most of the time, recreating a scene is very expensive, therefore,
        controls are asked to implement a reverse operation to undo their
        changes. Controls that need to store state in order to undo their
        actions should add data to the `target` object they received.

        Parameters
        ----------
        context : Dict[str, Any]
            The description of the scene to render
        """
        raise NotImplementedError



"""
threedb.result_logging.image_logger
====================================

Subclass of :mod:`threedb.result_logging.base_logger.BaseLogger`.
"""

import importlib
from os import makedirs, path
from typing import Any, Dict, Type

import cv2
import numpy as np
from threedb.rendering.base_renderer import BaseRenderer
from threedb.result_logging.base_logger import BaseLogger
from threedb.utils import CyclicBuffer

class ImageLogger(BaseLogger):
    """
    This logger saves all the images generated by 3DB during a given experiment.

    The renderer should expose a static property ``KEYS`` (e.g.
    :attr:`threedb.rendering.render_blender.Blender.KEYS`) that contains the
    dictionary keys that correspond to renderered images. For each key, images
    of that type are saved to an ``images/ID_<KEY>`` subdirectory of the given
    logging directory.
    """

    def __init__(self,
                 save_dir: str,
                 result_buffer: CyclicBuffer,
                 config: Dict[str, Any]) -> None:
        """Initializes an ImageLogger

        Parameters
        ----------
        save_dir : str
            Parent directory in which the images will be saved (in an
            ``images/`` subdirectory)
        result_buffer : CyclicBuffer
            The buffer that is being written to by the policy controller
            containing all the results
        config : Dict[str, Any]
            Config that the experiment was run with (see `here 
            <quickstart.html#writing-a-configuration-file>`_ for information
            on config file format)
        """
        super().__init__(save_dir, result_buffer, config)
        engine = importlib.import_module(config['render_args']['engine'])
        rendering_module: Type[BaseRenderer] = getattr(engine, 'Renderer')
        self.save_keys = rendering_module.KEYS
        self.result_buffer = result_buffer
        self.regid = self.result_buffer.register()

        self.folder = path.join(save_dir, 'images/')
        if not path.exists(self.folder):
            makedirs(self.folder)
        print(f'==> [Logging images to {self.folder} with regid {self.regid}]')

    def log(self, item: Dict[str, Any]) -> None:
        """Implementation of ``log()`` for ImageLogger

        Parameters
        ----------
        item : Dict[str, Any]
            Writes the images (as dictated by the
            :attr:`threedb.rendering.base_renderer.BaseRenderer.KEYS`
            field in the renderer) to the appropriate subdirectory.
        """
        rix = item['result_ix']
        buf_data = self.result_buffer[rix]
        for channel_name in self.save_keys:
            if not channel_name in buf_data.keys():
                continue

            image = buf_data[channel_name]
            if channel_name == 'segmentation':
                img_name = item['id'] + '_' + channel_name + '.npy'
                img_path = path.join(self.folder, img_name)
                np.save(img_path, image.numpy()[0])
            else:
                img_name = item['id'] + '_' + channel_name + '.png'
                img_path = path.join(self.folder, img_name)
                img_arr = image.permute(1, 2, 0).numpy() * 255.0
                img_to_write = cv2.cvtColor(img_arr, cv2.COLOR_RGB2BGR)
                cv2.imwrite(img_path, img_to_write)
        self.result_buffer.free(rix, self.regid)
    
    def end(self) -> None:
        pass

Logger = ImageLogger



"""
threedb.utils
=============
"""

import importlib
import ssl
from copy import deepcopy
from multiprocessing import Queue
from queue import Empty
from threedb.controls.base_control import BaseControl
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import numpy as np
import torch as ch
from torch import Tensor
from torch.types import _dtype
from torchvision import transforms
from tqdm import tqdm

def str_to_dtype(dtype_str: str) -> _dtype:
    return getattr(ch, dtype_str)

class CyclicBuffer:
    """
    A concurrent cyclic buffer with reference counting to store the result
    and avoid copying them to every sub process.


    """
    def __init__(self, buffers: Optional[Dict[str, Tuple[List[int], str]]] = None,
                 size: int = 1001, with_tqdm: bool = True) -> None:
        self.buffers = {}
        if buffers is None:
            buffers = {}
            self.initialized = False
        else:
            self.declare_buffers(buffers)

        self.used_buffer = np.zeros(size, dtype='uint8')
        self._free_idx = list(range(size))
        self.size = size

        self.first = 0
        self.last = 0
        self.mask = 0
        self.registration_count = 0
        self.events = Queue()

        self.progress_bar = None
        if with_tqdm:
            self.progress_bar = tqdm(unit='slots', desc='Buffer left',
                                     total=self.size, smoothing=0)

    def declare_buffers(self, buffers: Dict[str, Tuple[List[int], str]]) -> bool:
        if self.initialized and (buffers != self.declared_buffers):
            return False
        elif self.initialized:
            return True
        
        self.initialized = True
        self.declared_buffers = buffers
        for buf_name, (buf_size, buf_dtype) in buffers.items():
            buf = ch.zeros((self.size, *buf_size), dtype=str_to_dtype(buf_dtype)).share_memory_()
            self.buffers[buf_name] = buf
        return True

    def __getitem__(self, ind: int) -> Dict[str, ch.Tensor]:
        assert self.initialized, 'Buffer has not been initialized'
        return {k: v[ind] for (k, v) in self.buffers.items()}

    def free(self, ind: int, reg_id: int):
        assert self.initialized, 'Buffer has not been initialized'
        self.events.put((ind, reg_id))

    def register(self):
        self.registration_count += 1
        if self.registration_count > 8:
            raise Exception("Too many registrations")
        self.mask = 2**self.registration_count - 1
        return self.registration_count

    def process_events(self):
        assert self.initialized, 'Buffer has not been initialized'
        while True:
            try:
                (event, reg_id) = self.events.get(block=False)
                assert self.used_buffer[event] > 0
                if reg_id == -1:
                    self.used_buffer[event] = 0
                else:
                    self.used_buffer[event] ^= 1 << (reg_id - 1)
                if self.used_buffer[event] == 0:
                    self._free_idx.append(event)
                    if self.progress_bar is not None:
                        self.progress_bar.update(-1)
            except Empty:
                break

    def next_find_index(self) -> int:
        assert self.initialized, 'Buffer has not been initialized'
        while True:
            self.process_events()
            try:
                ind = self._free_idx.pop()
                if self.progress_bar is not None:
                    self.progress_bar.update(1)
                assert self.used_buffer[ind] == 0
                self.used_buffer[ind] = self.mask
                return ind
            except IndexError:
                # Should we add some kind of logging here?
                np.save('/tmp/used_buffer.npy', self.used_buffer)

    def allocate(self, data: Dict[str, ch.Tensor]):
        assert self.initialized, 'Buffer has not been initialized'
        next_ind = self.next_find_index()

        for buf_key, buf_data in data.items():
            assert buf_key in self.buffers, "Unexpected channel " + buf_key
            assert buf_data.dtype == self.buffers[buf_key].dtype, \
                f"Expected datatype {self.buffers[buf_key].dtype}, got {buf_data.dtype} for key {buf_key}"
            self.buffers[buf_key][next_ind] = buf_data

        return next_ind

    def close(self):
        if self.progress_bar is not None:
            self.progress_bar.close()

def overwrite_control(control: BaseControl, data: Dict[str, Union[Tuple[float, float], List[Any]]]):
    for key, val in data.items():
        if key in control.continuous_dims:
            control.update_continuous_dim(key, cast(Tuple[float, float], val))
        elif key in control.discrete_dims:
            control.update_discrete_dim(key, cast(List[Any], val))
        else:
            raise AttributeError(
                f"Attribute {key} unknown in {type(control).__name__}")

def init_control(cfg: Dict[str, Any], root_folder: str):
    args = cfg.get('args', {})
    module = importlib.import_module(cfg['module'])

    control_module = getattr(module, f"Control")
    control = control_module(**args, root_folder=root_folder)
    filtered_desc = {k: v for (k, v) in cfg.items() if k not in ['args', 'module']}
    overwrite_control(control, filtered_desc)
    return control


def init_policy(description):
    module = importlib.import_module(description['module'])
    return module.Policy(**{k: v for (k, v) in description.items() if k != 'module'})


def load_inference_model(args):
    try:
        _create_unverified_https_context = ssl._create_unverified_context
    except AttributeError:
        pass
    else:
        previous_context = ssl._create_default_https_context
        ssl._create_default_https_context = _create_unverified_https_context


    loaded_module = importlib.import_module(args['module'])
    model_args = args['args']

    model = getattr(loaded_module, args['class'])(**model_args)
    model.eval()

    ssl._create_default_https_context = previous_context

    def resize(tens):
        return ch.nn.functional.interpolate(tens[None], size=args['resolution'], mode='bilinear')[0]

    my_preprocess = transforms.Compose([
        resize,
        transforms.Normalize(mean=args['normalization']['mean'],
                             std=args['normalization']['std'])
    ])

    def inference_function(image):
        image = my_preprocess(image)
        return model(image.unsqueeze(0))[0], image.shape

    return inference_function



"""
threedb.controls.blender.denoiser
=================================

Denoise renderings. An example config file using this control can be found here:
`<https://github.com/3db/3db/tree/main/examples/unit_tests/denoiser.yaml>`_. (And
of the other controls in the ``unit_tests`` folder, this files parent folder, use
this control as well for more realistic renderings).

"""

from typing import Any, Dict
from ...try_bpy import bpy
from threedb.controls.base_control import PreProcessControl


class DenoiseControl(PreProcessControl):
    """Enable the built-in Denoise feature in blender

    Note
    ----
    This control enables the blender feature for denoising renderings.
    It has no parameters.

    .. admonition:: Example images

        .. thumbnail:: /_static/logs/denoiser/images/image_1.png
            :width: 100
            :group: denoiser

        .. thumbnail:: /_static/logs/denoiser/images/image_2.png
            :width: 100
            :group: denoiser

        .. thumbnail:: /_static/logs/denoiser/images/image_3.png
            :width: 100
            :group: denoiser

        .. thumbnail:: /_static/logs/denoiser/images/image_4.png
            :width: 100
            :group: denoiser

        .. thumbnail:: /_static/logs/denoiser/images/image_5.png
            :width: 100
            :group: denoiser

        With denoising.

    .. admonition:: Example images

        .. thumbnail:: /_static/logs/no_denoiser/images/image_1.png
            :width: 100
            :group: no_denoiser

        .. thumbnail:: /_static/logs/no_denoiser/images/image_2.png
            :width: 100
            :group: no_denoiser

        .. thumbnail:: /_static/logs/no_denoiser/images/image_3.png
            :width: 100
            :group: no_denoiser

        .. thumbnail:: /_static/logs/no_denoiser/images/image_4.png
            :width: 100
            :group: no_denoiser

        .. thumbnail:: /_static/logs/no_denoiser/images/image_5.png
            :width: 100
            :group: no_denoiser

        Without denoising.
    """

    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        """Enable OPENIMAGEDENOISE denoiser

        Parameters
        ----------
        context : Dict[str, Any]
            The scene context
        """
        bpy.context.scene.cycles.use_denoising = True
        bpy.context.scene.cycles.denoiser = 'OPENIMAGEDENOISE'
    
    def unapply(self, context: Dict[str, Any]) -> None:
        bpy.context.scene.cycles.use_denoising = False

Control = DenoiseControl



"""
threedb.scheduling.search_space
===============================

This file includes the `SearchSpace` class which given a list of control and their argements, 
creates the corresponding search space of the cross product of all controls.
"""

from ast import literal_eval
import numpy as np


class SearchSpace:

    def __init__(self, controls):
        self.controls = controls

        continuous_args = []
        discrete_args = {}
        set_args = []

        for control in self.controls:
            tpe = type(control)
            name = f"{tpe.__name__}"
            for continous_arg, value_range in control.continuous_dims.items():
                if isinstance(value_range, str):
                    value_range = literal_eval(value_range)
                try:
                    if len(value_range) != 2:
                        raise ValueError(
                            'range {value_range} for {control} should have length 2')
                    continuous_args.append((name, continous_arg, value_range))
                except TypeError:
                    # This is not a range but a constant value
                    # We consider this paramter to be set and not searched over
                    set_args.append((name, continous_arg, value_range))
            for k, v in control.discrete_dims.items():
                print(v)
                if not isinstance(v, (tuple, list)):
                    v = [v]
                if len(v) == 1:
                    set_args.append((name, k, v[0]))
                else:
                    discrete_args[(name, k)] = v

        self.continuous_args = continuous_args
        self.discrete_args = discrete_args
        self.set_args = set_args
        print(discrete_args)

    def generate_description(self):
        return len(self.continuous_args), [len(x) for x in self.discrete_args.values()]

    def generate_log(self, packed_continuous, packed_discrete):
        pass

    def unpack(self, packed_continuous, packed_discrete):
        result = {}
        for (control_name, attr_name, (start, end)), value in zip(self.continuous_args, packed_continuous):
            result[(control_name, attr_name)] = value * (end - start) + start

        for ((control_name, attr_name), values) , ix in zip(self.discrete_args.items(), packed_discrete):
            result[(control_name, attr_name)] = values[ix]

        for (control_name, attr_name, value) in self.set_args:
            result[(control_name, attr_name)] = value

        order_controls = [(type(x).__module__, type(x).__name__) for x in self.controls]

        return result, order_controls



"""
threedb.rendering.base_renderer
===============================

An abstract class of a Renderer which must be subclassed when adding new renderers.
See `here <custom_renderer.html>`_ for details.
"""

import torch as ch
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional, Tuple
from threedb.rendering.utils import ControlsApplier
from typing import List

class RenderEnv(ABC): pass
class RenderObject(ABC): pass

class BaseRenderer(ABC):
    NAME: str = 'BaseRenderer'
    KEYS: List[str] = ['rgb']

    def __init__(self, 
                 root_dir: str, 
                 render_settings: Dict[str, Any],
                 env_extensions: List[str] = []) -> None:
        self.root_dir = root_dir
        self.env_extensions = env_extensions
        self.args = render_settings
    
    @staticmethod
    @abstractmethod
    def enumerate_models(search_dir: str) -> List[str]:
        """
        Given a root folder, returns all valid .blend files (according to a
        per-renderer convention).
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def enumerate_environments(search_dir: str) -> List[str]:
        """
        Given a root folder, returns all files in root/blender_environments/ which
        have extensions in ENV_EXTENSIONS above.
        """
        raise NotImplementedError

    @abstractmethod
    def declare_outputs(self) -> Dict[str, Tuple[List[int], str]]:
        """
        This function declares what the output of render() will be, based on the
        renderer settings. Returns a dictionary mapping keys to (dtype, size)
        tuples---the output of render() is string-to-tensor dictionary whose
        tensors will be checked against the return value of this function for
        both size and type.

        A basic implementation which suffices for most applications is provided
        in the abstract class :class:`threedb.rendering.base_renderer.BaseRenderer`.
        """
        raise NotImplementedError

    @abstractmethod
    def load_model(self, model: str) -> RenderObject:
        """
        Given a root directory and a model id, loads the model into the renderer
        and returns the corresponding object.
        """
        raise NotImplementedError
    
    @abstractmethod
    def get_model_uid(self, model: RenderObject) -> str:
        """
        Given an 3D model, return its UID as assigned by the renderer.
        """

    @abstractmethod
    def load_env(self, env: str) -> Optional[RenderEnv]:
        """
        Given a root folder and environment ID, load the environment into
        the renderer. If needed, returns an environment object, to be passed
        back to the render() function.
        """
        raise NotImplementedError

    @abstractmethod
    def setup_render(self, 
                     model: Optional[RenderObject], 
                     env: Optional[RenderEnv]) -> None:
        """
        Perform setup operations for rendering. Called only when the model or
        environment being rendered changes---otherwise, only render() will be
        called. No return value.
        """
        raise NotImplementedError
    
    @abstractmethod
    def get_context_dict(self, model_uid: str, object_class: int) -> Dict[str, Any]:
        """Returns, for a given model id, a "context" dictionary; this context
        will be passed to the controls (see :class::class:`threedb.controls.base_control.PreProcessControl` for more info).

        Parameters
        ----------
        model_uid : str
            The ID of the model being rendered.
        object_class : int
            The class label for the model.

        Returns
        -------
        Dict[str, Any]
            A context dictionary 
        """

    @abstractmethod
    def render(self, 
               model_uid: str, 
               loaded_model: RenderObject, 
               loaded_env: RenderEnv) -> Dict[str, ch.Tensor]:
        """[summary]

        Parameters
        ----------
        model_uid : str
            Render a model and environment. You can assume that setup_render() has
            been called with the relevant model and object in context. This function
            should also handle applying the pre and post-processing controls.
        loaded_model : RenderObject
            The model that was most recently loaded and passed to setup_render.
        loaded_env : RenderEnv
            the environment that was most recently loaded and passed to setup_render. 

        Returns
        -------
        Dict[str, ch.Tensor]
            A dictionary mapping result keys (e.g., 'rgb', 'segmentation', etc.)
            to PyTorch tensor outputs.
        """        
        raise NotImplementedError


"""
threedb.controls.blender.scale
==============================

Control the object scale. An example config file using this control can be found here:
`<https://github.com/3db/3db/tree/main/examples/unit_tests/scale.yaml>`_.
"""

from typing import Any, Dict
from threedb.controls.base_control import PreProcessControl

class ObjScaleControl(PreProcessControl):
    """This control scales the object.

    Continuous Parameters:

    - ``factor``: scaling factor which takes any positive number. Setting the
        factor to 1 maintains the same object size. (default range: ``[0.25, 1]``)

    .. admonition:: Example images

        .. thumbnail:: /_static/logs/scale/images/image_1.png
            :width: 100
            :group: scale

        .. thumbnail:: /_static/logs/scale/images/image_2.png
            :width: 100
            :group: scale

        .. thumbnail:: /_static/logs/scale/images/image_3.png
            :width: 100
            :group: scale

        .. thumbnail:: /_static/logs/scale/images/image_4.png
            :width: 100
            :group: scale

        .. thumbnail:: /_static/logs/scale/images/image_5.png
            :width: 100
            :group: scale
        
        Examples of scale at different levels.
    """
    def __init__(self, root_folder: str):
        continuous_dims = {
            'factor': (0.25, 1.),
        }
        super().__init__(root_folder, continuous_dims=continuous_dims)

    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        """scales the object in the scene context by a factor `factor`

        Parameters
        ----------
        context
            The scene context object
        factor
            Scaling factor which takes any positive number. Setting the
            factor to 1 maintains the same object size.
        """
        self.ob = context['object']
        self.ob.scale = (control_args['factor'],) * 3

    def unapply(self, context):
        """Rescales the object to its original dimensions

        Parameters
        ----------
        context
            The scene context object
        """
        self.ob.scale = (1.,) * 3

Control = ObjScaleControl



"""
Rendering Utils
===============

Some useful untilities for Blender.
"""

from typing import Any, Dict, List, Tuple
import numpy as np
from collections import defaultdict
import importlib
from ..try_bpy import mathutils
from ..controls.base_control import PreProcessControl, PostProcessControl

def sample_upper_sphere() -> mathutils.Vector:
    vec = np.random.randn(3)
    vec /= np.linalg.norm(vec)
    vec[2] = np.abs(vec[2])

    return mathutils.Vector(vec)

def lookat_viewport(target: mathutils.Vector, 
                    location: mathutils.Vector) -> mathutils.Vector:
    diff = location - target
    diff = diff.normalized()
    rot_z = (np.arctan(diff.y / diff.x))
    if diff.x > 0:
        rot_z += np.pi / 2
    else:
        rot_z -= np.pi / 2
    return mathutils.Euler((np.arccos(diff[2]), 0, rot_z)).to_quaternion()


class ControlsApplier:

    def __init__(self, control_list: List[Tuple[str, str]], 
                       render_args: Dict[Tuple[str, str], Any],
                       controls_args: Dict[str, Dict[str, Any]],
                       root_folder: str):
        control_classes = []

        for module, classname in control_list:
            imported = importlib.import_module(module)
            control_classes.append(
                getattr(imported, classname)(
                    root_folder=root_folder,
                    **controls_args[classname])
            )

        grouped_args = defaultdict(dict)

        for (classname, attribute), value in render_args.items():
            grouped_args[classname][attribute] = value

        self.control_classes = control_classes
        self.grouped_args = grouped_args

    def apply_pre_controls(self, context: Dict[str, Any]) -> None:
        for control_class in self.control_classes:
            if isinstance(control_class, PreProcessControl):
                classname = type(control_class).__name__
                control_params = self.grouped_args[classname]
                control_class.apply(context=context, control_args=control_params)

    # Unapply controls (e.g. delete occlusion objects, rescale object, etc)
    def unapply(self, context: Dict[str, Any]) -> None:
        for control_class in self.control_classes:
            if isinstance(control_class, PreProcessControl):
                control_class.unapply(context)

    # post-processing controls
    def apply_post_controls(self, img):
        for control_class in self.control_classes:
            if isinstance(control_class, PostProcessControl):
                classname = type(control_class).__name__
                control_params = self.grouped_args[classname]
                img = control_class.apply(render=img, control_args=control_params)
        return img



"""
threedb.policies.grid_search
============================

A search policy over controls that enumerates the cross product of all controls 
and sweep over them all.
"""
import numpy as np
from itertools import product

def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


class GridSearchPolicy:
    def __init__(self, continuous_dim, discrete_sizes, samples_per_dim):
        self.continuous_dim = continuous_dim
        self.discrete_sizes = discrete_sizes
        self.samples_per_dim = samples_per_dim

    def hint_scheduler(self):
        total_queries = self.samples_per_dim ** self.continuous_dim
        total_queries *= np.prod(self.discrete_sizes)
        return 1, int(total_queries)

    def run(self, render_and_send):
        continuous_values = np.linspace(0, 1, self.samples_per_dim)

        discrete_spaces = []

        for n in self.discrete_sizes:
            discrete_spaces.append(np.arange(n))

        result = []
        for continuous_instance in product(*([continuous_values] * self.continuous_dim)):
            for discrete_instance in product(*discrete_spaces):
                result.append((continuous_instance, discrete_instance))

        for r in chunks(result, 1000):
            render_and_send(r)

Policy = GridSearchPolicy



"""
threedb.controls.blender.pin_to_ground
======================================

Control where the bottom bounding box of an object lies on the z axis. An example
config file using this control can be found here:
`<https://github.com/3db/3db/tree/main/examples/unit_tests/pin_to_ground.yaml>`_.
"""
from typing import Any, Dict
from ...try_bpy import bpy, mathutils
from threedb.controls.base_control import PreProcessControl
from .utils import cleanup_translate_containers, post_translate


class PinToGroundControl(PreProcessControl):
    """Control that moves an object vertically to touch the ground
    Useful when you want a slightly more realistic rendering 
    (i.e. avoid flying objects)

    Continuous Dimensions:

    - ``z_ground``: the z-coordinate that the bottom of the object's bounding
       moves to. Range: ``[0, 1]``.

    .. note::
        This control come after the ``PositionControl`` and
        ``OrientationControl`` controls. In these patterns, you first move the object to a location
        of interest, then drag it to the ground under that location.

    .. admonition:: Example images

        .. thumbnail:: /_static/logs/pin_to_ground/images/image_1.png
            :width: 100
            :group: pin_to_ground

        .. thumbnail:: /_static/logs/pin_to_ground/images/image_2.png
            :width: 100
            :group: pin_to_ground

        .. thumbnail:: /_static/logs/pin_to_ground/images/image_3.png
            :width: 100
            :group: pin_to_ground

        .. thumbnail:: /_static/logs/pin_to_ground/images/image_4.png
            :width: 100
            :group: pin_to_ground

        .. thumbnail:: /_static/logs/pin_to_ground/images/image_5.png
            :width: 100
            :group: pin_to_ground

        Varying ``z_ground`` across its range.
    """

    def __init__(self, root_folder: str):
        continuous_dims = {
            'z_ground': (0., 1.),
        }
        super().__init__(root_folder, continuous_dims=continuous_dims)

    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        """Pins the object to the ground

        Parameters
        ----------
        context : Dict[str, Any]
            The scene context object
        control_args : Dict[str, Any]
            The parameters for this control; should have key ``z_ground``
            mapping to any real number, see class docstring for documentation.
        """
        no_err, msg = self.check_arguments(control_args)
        assert no_err, msg

        obj = context["object"]
        bpy.context.view_layer.update()
        obj_min_z = min((obj.matrix_world @ x.co)[2] for x in obj.data.vertices)
        post_translate(obj, mathutils.Vector([0, 0, control_args['z_ground'] - obj_min_z]))

    def unapply(self, context):
        cleanup_translate_containers(context['object'])

Control = PinToGroundControl



"""
threedb.policies.random_search
==============================

A search policy over controls that randomly samples from the set of 
permissible controls.
"""

import numpy as np
from itertools import product

class RandomSearchPolicy:
    def __init__(self, continuous_dim, discrete_sizes, samples, seed=None):
        """
            Pick a total number of samples randomly
        """
        self.continuous_dim = continuous_dim
        self.discrete_sizes = discrete_sizes
        self.samples = samples
        self.seed = seed

    def hint_scheduler(self):
        return 1, self.samples 

    def run(self, render_and_send):
        rng = np.random.default_rng(self.seed)
        result = []
        for _ in range(self.samples):
            continuous_instance = rng.random(self.continuous_dim)
            discrete_instance = [rng.integers(low=0, high=n) for n in self.discrete_sizes]
            result.append((continuous_instance, discrete_instance))

        render_and_send(result)

Policy = RandomSearchPolicy


"""
threedb.result_logging.logger_manager
=====================================

"""

from threedb.result_logging.base_logger import BaseLogger

class LoggerManager():
    """
    A LoggerManager allows us to log from several loggers at once, without
    handling each one individually.

    Users should not have to modify or subclass this to extend 3DB.
    """
    def __init__(self):
        super().__init__()
        self.loggers = []

    def append(self, logger: BaseLogger):
        """
        Adds a new logger.
        """
        self.loggers.append(logger)

    def log(self, item):
        """
        Logs the given items from each logger under management.
        """
        for logger in self.loggers:
            logger.enqueue(item)

    def start(self):
        """
        Starts each logger under management.
        """
        for logger in self.loggers:
            logger.start()

    def join(self):
        """
        Wait for each logger under management to complete their tasks.
        """
        for logger in self.loggers:
            logger.join()


"""
threedb.result_logging.tb_logger
==================================

Subclass of :mod:`threedb.result_logging.base_logger.BaseLogger`.
"""

import pandas as pd

from threedb.result_logging.base_logger import BaseLogger
from threedb.utils import CyclicBuffer
from torch.utils.tensorboard import SummaryWriter

class TbLogger(BaseLogger):
    """
    TensorBoard Logger.

    Logs accuracy (as a scalar) and all generated images to TensorBoard.
    """
    def __init__(self, tb_dir: str, result_buffer: CyclicBuffer, _) -> None:
        super().__init__(tb_dir, result_buffer, None)
        self.tb_dir = tb_dir
        print(f'==> [Logging tensorboard to {tb_dir}]')
        self.writer = None # Defer allocation in the sub-process
        self.result_buffer = result_buffer
        self.regid = self.result_buffer.register()
        self.numeric_data = []
        self.images = {}
        self.count = 0

    def _write(self):
        """
        Internal function for writing to the tensorboard.
        """
        if self.writer is None:
            self.writer = SummaryWriter(log_dir=self.tb_dir)
        data_df = pd.DataFrame(self.numeric_data)
        current_acc = data_df.is_correct.mean()
        self.writer.add_scalar('Accuracy', current_acc, self.count)
        for uid in data_df.model.unique():
            image_id = data_df[data_df.model == uid].id.iloc[-1]
            self.writer.add_image(uid, self.images[image_id], self.count)
        self.images = {}
        self.numeric_data = []

    def log(self, item):
        self.count += 1
        rix = item['result_ix']
        buf_data = self.result_buffer[rix]
        print(buf_data.keys(), item.keys())
        information = {k: v for k, v in item.items() if k != 'result_ix'}
        information['is_correct'] = buf_data

        self.numeric_data.append(information)
        self.images[item['id']] = image['rgb'].clone()

        if self.count % 1 == 0:
            self._write()

        self.result_buffer.free(rix, self.regid)
    
    def end(self) -> None:
        pass

Logger = TbLogger


"""
threedb.result_logging.json_logger
==================================

Subclass of :mod:`threedb.result_logging.base_logger.BaseLogger`.
"""

import copy
import importlib
import shutil
from os import path
from typing import Any, Dict, Iterable, Tuple

import numpy as np
import orjson as json
import torch as ch
from threedb.result_logging.base_logger import BaseLogger
from threedb.utils import CyclicBuffer

def clean_key(k: Iterable[str]) -> str:
    """
    Utility function for formatting keys.
    This is a no-op if the input is a string, otherwise expects an iterable
    of strings, which it joins with a period.
    """
    return k if isinstance(k, str) else '.'.join(k)

def clean_value(val: Any):
    """
    Utility function for formatting tensors.
    Converts torch tensors to numpy, is a no-op for all other types.
    """
    return val.numpy() if ch.is_tensor(val) else val

def json_default(obj: np.ndarray) -> str:
    """
    Another utility function; turns floats into strings, otherwise (if the
    input does not have type ``np.float64`` raises a ``TypeError``.
    """
    if isinstance(obj, np.ndarray):
        return str(obj)
    raise TypeError

def clean_log(log_d: dict, key_blacklist: Tuple[str, str] = ('image', 'result_ix')) -> dict:
    """
    Cleans a dictionary for log-writing. In particular, all keys (expected to
    be either strings or iterables of strings) are converted to strings, and
    all torch tensors are converted to numpy arrays.
    """
    cleaned = {}
    for k, val in log_d.items():
        if k in key_blacklist:
            continue
        clean_v = clean_log(val) if isinstance(val, dict) else clean_value(val)
        cleaned[clean_key(k)] = clean_v
    return cleaned


class JSONLogger(BaseLogger):
    def __init__(self,
                 root_dir: str,
                 result_buffer: CyclicBuffer,
                 config: Dict[str, Dict[str, Any]]) -> None:
        """
        A logger that logs all an experiments meta-data and results into a JSON file.
        """
        super().__init__(root_dir, result_buffer, config)
        fname = path.join(root_dir, 'details.log')
        self.handle = open(fname, 'ab+')
        self.regid = self.buffer.register()
        self.evaluator = importlib.import_module(self.config['evaluation']['module']).Evaluator
        if 'label_map' in config['inference']:
            classmap_fname = path.join(root_dir, 'class_maps.json')
            print(f"==> [Saving class maps to {classmap_fname}]")
            shutil.copyfile(config['inference']['label_map'], classmap_fname)
        print(f'==> [Logging to the JSON file {fname} with regid {self.regid}]')

    def log(self, item: Dict[str, Any]) -> None:
        """Concrete implementation of
        :meth:`threedb.result_logging.base_logger.BaseLogger.log`.

        Parameters
        ----------
        item : Dict[str, Any]
            The item to be logged.
        """
        item = copy.deepcopy(item)
        rix = item['result_ix']
        buffer_data = self.buffer[rix]
        result = {k: v for (k, v) in buffer_data.items() if k in self.evaluator.KEYS}
        for k in ['id', 'environment', 'model', 'render_args']:
            result[k] = item[k]
        result['output_type'] = self.evaluator.output_type
        cleaned = clean_log(result)
        encoded = json.dumps(cleaned, default=json_default,
                             option=json.OPT_SERIALIZE_NUMPY | json.OPT_APPEND_NEWLINE)
        self.buffer.free(rix, self.regid)
        self.handle.write(encoded)

    def end(self):
        """Concrete implementation of
        :meth:`threedb.result_logging.base_logger.BaseLogger.end`. 
        
        Closes the necessary file handle.
        """
        self.handle.close()

Logger = JSONLogger



"""
threedb.controls.blender.camera
===============================

Control the camera. An example config file using this control can be found here:
`<https://github.com/3db/3db/tree/main/examples/unit_tests/camera.yaml>`_.
"""

from typing import Any, Dict

import copy
from ...try_bpy import bpy, mathutils 
from ..base_control import PreProcessControl
from ...rendering.utils import lookat_viewport


class CameraControl(PreProcessControl):
    """Control that changes the camera that will be used to render the image

    Continuous Dimensions:

    - ``view_point_x``: The original x coordinate of the camera (see the note
      below). (range: ``[-1, 1]``)
    - ``view_point_y``: The original y coordinate of the camera (see the note
      below). (range: ``[-1, 1]``)
    - ``view_point_z``: The original z coordinate of the camera (see the note
      below). (range: ``[0, 1]``)
    - ``zoom_factor``: Defines how much should we see of the object. A
      ``zoom_factor`` of 1 means we completely see the object with a little
      margin. above 1 we are close. 
      below 1 we are further. (range: ``[0.5, 2]``)
    - ``aperture``: The aperture of the camera. (range: ``[1, 32]``)
    - ``focal_length``: The focal length of the camera. (range: ``[10, 400]``)

    .. admonition:: Example images

        .. thumbnail:: /_static/logs/camera/images/image_1.png
            :width: 100
            :group: camera

        .. thumbnail:: /_static/logs/camera/images/image_2.png
            :width: 100
            :group: camera

        .. thumbnail:: /_static/logs/camera/images/image_3.png
            :width: 100
            :group: camera

        .. thumbnail:: /_static/logs/camera/images/image_4.png
            :width: 100
            :group: camera

        .. thumbnail:: /_static/logs/camera/images/image_5.png
            :width: 100
            :group: camera

        Varying each parameter across its range.

    .. note::
        The camera uses the parameters as following:

        1. We set the ``aperture`` and ``focal_length``
        2. We move the camera according to ``view_point_{x,y,z}`` and look at
           the object
        3. We move closer or further in order to satisfy the ``zoom_factor``
           constraint

        In most cases the final position of the camera will be
        quite different from the ``view_point`` parameter.
    """
    def __init__(self, root_folder: str):
        continuous_dims = {
            'view_point_x': (-1, 1),
            'view_point_y': (-1, 1),
            'view_point_z': (0, 1),
            'zoom_factor': (0.5, 2),
            'aperture': (1, 32),
            'focal_length': (10, 400),
        }
        super().__init__(root_folder, continuous_dims=continuous_dims)

    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        no_err, msg = self.check_arguments(control_args)
        assert no_err, msg

        args = copy.copy(control_args)
        zoomout_factor = 1 / args['zoom_factor']

        camera = bpy.data.objects['Camera']
        camera.data.lens = args['focal_length']
        camera.data.dof.aperture_fstop = args['aperture']
        camera.data.clip_start = 0.001

        obj = context['object']

        # avoid division by zero bug
        args['view_point_x'] += 1e-6
        args['view_point_y'] += 1e-6
        args['view_point_z'] += 1e-6

        bpy.ops.object.select_all(action='DESELECT')
        obj.select_set(True)
        for area in bpy.context.screen.areas:
            if area.type == 'VIEW_3D':
                ctx = bpy.context.copy()
                space = area.spaces[0]
                reg = space.region_3d
                ctx['area'] = area
                ctx['region'] = area.regions[-1]
                ctx['space_data'] = space
                ctx['camera'] = camera
                reg.view_location = obj.location
                reg.view_distance = 2
                previous_lens = space.lens
                space.lens = camera.data.lens * zoomout_factor
                v_dir = mathutils.Vector((args['view_point_x'], args['view_point_y'], args['view_point_z']))
                reg.view_rotation = lookat_viewport(reg.view_location,
                                                    reg.view_location + v_dir)
                bpy.ops.view3d.view_selected(ctx)
                bpy.ops.view3d.camera_to_view(ctx)
                space.lens = previous_lens
                camera.data.dof.focus_object = obj

    def unapply(self, context: Dict[str, Any]) -> None:
        pass

Control = CameraControl



"""
Rendering utils
===============

Common utils function for blender controls
"""
from os import path
from typing import Any, Tuple
from ...try_bpy import bpy

def load_model(model: str) -> str:
    """Load an object from a blender file and insert it in the scene

    Note
    ----
    Since .blend files can contain many object we go under the assumption
    that in the file `XXXXXX.blend`, there will be an object named `XXXXXX`,
    and this is this one we will load

    Parameters
    ----------
    model
        The path to the blender file that needs to be loaded

    Returns
    -------
    str
        The uid of the object loaded
    """
    basename, filename = path.split(model)
    uid = filename.replace('.blend', '')
    blendfile = path.join(basename, uid + '.blend')
    section = "\\Object\\"
    object = uid

    filepath = uid + '.blend'
    directory = blendfile + section
    filename = object

    bpy.ops.wm.append(
        filepath=filepath,
        filename=filename,
        directory=directory)

    return uid


TRANSLATE_PREFIX = 'translation_control_'


def cleanup_translate_containers(obj: Any) -> None:
    """Remove all translations containers

    Note
    ----
    To know what translate containers are please refer to `post_translate`

    Parameters
    ----------
    obj: blender object
        The object to remove containers on

    """
    obj.parent = None
    for other in bpy.data.objects:
        if other.type == 'EMPTY' and len(other.children) == 0:
            bpy.data.objects.remove(other, do_unlink=True)


def post_translate(obj: Any, offset: Tuple[float, float, float]) -> None:
    """Apply a translation on an object but ensure it happens after rotations

    Note
    ----
    To work this function creates a container around the object and apply
    the translation on the parent. This way, any rotation on the object
    itself will always be applied before the translation
    The container name will start by TRANSLATE_PREFIX so that is can easily
    be detected and removed in `cleanup_translate_containers`

    Parameters
    ----------
    obj: blender object
        The object to translate
    offset
        The vector to translate by
    """
    if (obj.parent is None
            or not obj.parent.name.startswith(TRANSLATE_PREFIX)):
        bpy.ops.object.empty_add(type='PLAIN_AXES', align='WORLD',
                                 location=(0, 0, 0), scale=(1, 1, 1))
        container = bpy.context.object
        container.name = TRANSLATE_PREFIX + obj.name
        obj.parent = container
    obj.parent.location += offset


def clamp(value: float, minimum: float, maximum: float) -> float:
    """Clamp a value between two numbers

    Parameters
    ----------
    value
        The value to clamp
    minimum
        Minimum acceptable value
    maximum
        Maximum acceptable value

    Returns
    -------
    float
        The clamped value
    """

    return max(minimum, min(value, maximum))


def camera_view_bounds_2d(scene: bpy.types.Scene,
                          cam_ob: bpy.types.Object,
                          mesh: bpy.types.Mesh) -> Tuple[float, float, float, float]:
    """Returns camera-space bounding box of mesh object.
    Negative 'z' value means the point is behind the camera.
    Takes shift-x/y, lens angle and sensor size into account
    as well as perspective/ortho projections.

    Parameters
    ----------
    scene : bpy.types.Scene
        Scene to use for frame size.
    cam_ob : bpy.types.Object
        Camera object
    me_ob : bpy.types.Mesh
        Untransformed Mesh.

    Returns
    -------
    Tuple[float, float, float, float]
        A tuple ``(x, y, width, height)`` encoding the camera-space bounding box
        of the mesh object.
    """    
    mat = cam_ob.matrix_world.normalized().inverted()
    depsgraph = bpy.context.evaluated_depsgraph_get()
    mesh_eval = mesh.evaluated_get(depsgraph)
    me = mesh_eval.to_mesh()
    me.transform(mesh.matrix_world)
    me.transform(mat)
    camera = cam_ob.data
    frame = [-v for v in camera.view_frame(scene=scene)[:3]]
    camera_persp = camera.type != "ORTHO"
    lx = []
    ly = []
    for v in me.vertices:
        co_local = v.co
        z = -co_local.z
        if camera_persp:
            if z == 0.0:
                lx.append(0.5)
                ly.append(0.5)
            # Does it make any sense to drop these?
            # if z <= 0.0:
            #    continue
            else:
                frame = [(v / (v.z / z)) for v in frame]
        min_x, max_x = frame[1].x, frame[2].x
        min_y, max_y = frame[0].y, frame[1].y
        x = (co_local.x - min_x) / (max_x - min_x)
        y = (co_local.y - min_y) / (max_y - min_y)
        lx.append(x)
        ly.append(y)

    min_x = clamp(min(lx), 0.0, 1.0)
    max_x = clamp(max(lx), 0.0, 1.0)
    min_y = clamp(min(ly), 0.0, 1.0)
    max_y = clamp(max(ly), 0.0, 1.0)
    mesh_eval.to_mesh_clear()
    r = scene.render
    fac = r.resolution_percentage * 0.01
    dim_x = r.resolution_x * fac
    dim_y = r.resolution_y * fac

    if round((max_x - min_x) * dim_x) == 0 or round((max_y - min_y) * dim_y) == 0:
        return (0, 0, 0, 0)
    return (
        round(min_x * dim_x),  # X
        round(dim_y - max_y * dim_y),  # Y
        round((max_x - min_x) * dim_x),  # Width
        round((max_y - min_y) * dim_y),  # Height
    )



"""
threedb.controls.blender.mug_liquid
===================================

Change liquid inside of a mug. Used to reproduce examples in the 3DB paper.
"""

from typing import Any, Dict
from ..base_control import PreProcessControl


class MugLiquidControl(PreProcessControl):
    """Change the material of the liquid present in the mug object

    Note
    ----
    This control assumes that the surface of the liquid has a manterial named
    "liquid" and that the object we want to modify is the target of this render.

    The ratio of water will be 1 - ratio_milk - ratio_coffee

    Continuous Dimensions:

    - ``ratio_milk`` (float): The ratio of milk (range: ``[0, 1]``)
    - ``ratio_coffee`` (float): The ratio of coffee (range: ``[0, 1]``)
    """

    def __init__(self, root_folder: str):
        continuous_dims = {
            'ratio_milk': (0, 1),
            'ratio_coffee': (0, 1),
        }

        super().__init__(root_folder, continuous_dims=continuous_dims)

    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        """Change the material of the liquid

        Parameters
        ----------
        context : Dict[str, Any]
            The scene context
        control_args : Dict[str, Any]
            The arguments for this control, should have keys ``ratio_milk`` and
            ``ratio_coffee`` (see the class docstring for information).
        """
        no_err, msg = self.check_arguments(control_args)
        assert no_err, msg

        ratio_coffee: float
        ratio_milk: float
        ratio_coffee, ratio_milk = control_args['ratio_coffee'], control_args['ratio_milk']

        ratio_water = 1 - ratio_coffee - ratio_milk

        obj = context['object']
        material = obj.material_slots['liquid'].material.node_tree
        material.nodes["coffee_milk_ratio"].outputs[0].default_value = (
            ratio_coffee / (ratio_coffee + ratio_milk))
        material.nodes["water_ratio"].outputs[0].default_value = ratio_water

    def unapply(self, context: Dict[str, Any]) -> None:
        pass


Control = MugLiquidControl



"""
threedb.controls.blender.orientation
====================================

Cover the subject object with another object. An example config file using this control can be found here:
`<https://github.com/3db/3db/tree/main/examples/unit_tests/orientation.yaml>`_. 
"""

from typing import Any, Dict
import numpy as np
from threedb.controls.base_control import PreProcessControl


class OrientationControl(PreProcessControl):
    """This control changes the orientation of the object (i.e. the rotation).

    Continuous Parameters:

    - ``rotation_x``: The x component of the Eulerian rotation (range: ``[-pi, pi]``)
    - ``rotation_y``: The y component of the Eulerian rotation (range: ``[-pi, pi]``)
    - ``rotation_z``: The z component of the Eulerian rotation (range: ``[-pi, pi]``)

    .. admonition:: Example images

        .. thumbnail:: /_static/logs/orientation/images/image_1.png
            :width: 100
            :group: orientation

        .. thumbnail:: /_static/logs/orientation/images/image_2.png
            :width: 100
            :group: orientation

        .. thumbnail:: /_static/logs/orientation/images/image_3.png
            :width: 100
            :group: orientation

        .. thumbnail:: /_static/logs/orientation/images/image_4.png
            :width: 100
            :group: orientation

        .. thumbnail:: /_static/logs/orientation/images/image_5.png
            :width: 100
            :group: orientation

        Varying the orientation across all parameters.
    """
    def __init__(self, root_folder: str):
        continuous_dims = {
            'rotation_x': (-np.pi, np.pi),
            'rotation_y': (-np.pi, np.pi),
            'rotation_z': (-np.pi, np.pi),
        }
        super().__init__(root_folder, continuous_dims=continuous_dims)

    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        """Rotates the object according to the given parameters

        Parameters
        ----------
        context : Dict[str, Any]
            The scene context object
        control_args : Dict[str, Any]
            The parameters for this control, ``rotation_x``, ``rotation_y``, and
            ``rotation_z``. See the class docstring for their documentation.
        """
        no_err, msg = self.check_arguments(control_args)
        assert no_err, msg

        obj = context['object']
        obj.rotation_mode = 'XYZ'
        obj.rotation_euler = (control_args['rotation_x'],
                              control_args['rotation_y'],
                              control_args['rotation_z'])

    def unapply(self, context: Dict[str, Any]) -> None:
        pass

Control = OrientationControl



"""
threedb.client
==============

The client is responsible for receiving tasks (sets of parameters) from the
main node, rendering these tasks, performing inference, and then returning
the results to the main node for inference.

Users should not have to modify or even view this file in order to use and
extend 3DB.
"""
import argparse
import importlib
import sys
import time
from types import SimpleNamespace
from typing import Any, Optional, Type, Dict
from uuid import uuid4

import numpy as np
import torch as ch
import zmq
from tqdm import tqdm

from threedb.rendering.utils import ControlsApplier
from threedb.rendering.base_renderer import BaseRenderer
from threedb.evaluators.base_evaluator import BaseEvaluator
from threedb.utils import load_inference_model

COUNTER = 0

def send_array(sock: zmq.Socket,
               arr: Any,
               dtype: Optional[str]=None,
               flags: int=0,
               copy: bool=True,
               track: bool=False):
    """send a numpy array with metadata"""
    if ch.is_tensor(arr):
        arr = arr.data.cpu().numpy()
    arr = np.ascontiguousarray(arr)
    if dtype is not None:
        arr = arr.astype(dtype)
    message_dict = dict(
        dtype=str(arr.dtype),
        shape=arr.shape,
    )
    sock.send_json(message_dict, flags | zmq.SNDMORE)
    return sock.send(arr, flags, copy=copy, track=track)

def query(sock: zmq.Socket, kind: str, worker_id: str, 
          result_data: Optional[Dict[str, Any]] = None,
          result_dtypes: Optional[Dict[str, str]] = None,
          **kwargs) -> Dict[str, Any]:
    """Send a request back to the server and receive a response. Additional
    named arguments are forwarded to the server as-is as part of the request.

    Parameters
    ----------
    sock : zmq.Socket
        An open socket connected to the same port as the server.
    kind : str
        What kind of request to send (``'info'``, ``'push'``, ``'pull'``, or
        ``'decl'``)
    worker_id : str
        The id of the client sending the request
    result_data : Optional[Dict[str, Any]], optional
        If ``kind == 'push'``, this should be a dictionary of results to send
        back to the server (otherwise ignored), by default None.

    Returns
    -------
    Dict[str, Any]
        The response from the server for to the sent message; see `here <extending.html#the-3db-workflow>`_ for
        documentation of the communication protocol,
    """            
    to_send = {
        'kind': kind,
        'worker_id': worker_id,
        **kwargs
    }

    if result_data is not None:
        result_keys = list(result_data.keys())
        to_send['result_keys'] = result_keys

        sock.send_json(to_send, flags=zmq.SNDMORE)
        for channel_name in result_keys:
            send_array(sock,
                       result_data[channel_name],
                       result_dtypes.get(channel_name, None),
                       flags=zmq.SNDMORE)
        sock.send_string('done')
    else:
        sock.send_json(to_send, flags=0)

    response = sock.recv_pyobj()
    if kind == 'decl':
        assert response['kind'] == 'ack', 'Received a non-ack message from the server, abort.'
        
    if response['kind'] == 'die':
        print("==> [Received closed request from master]")
        sys.exit()
    return response

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Render worker for the robustness threedb')
    parser.add_argument('root_folder', type=str,
                        help='folder containing all data (models, environments, etc)')
    parser.add_argument('--master-address', '-a', type=str,
                        help='How to contact the master node',
                        default='localhost:5555')
    parser.add_argument('--gpu-id', help='The GPU to use to render (-1 for cpu)',
                        default=-1, type=int,)
    parser.add_argument('--cpu-cores', help='number of cpu cores to use (default uses all)',
                        default=None, type=int,)
    parser.add_argument('--tile-size', help='The size of tiles used for GPU rendering',
                        default=32, type=int)
    parser.add_argument('--batch-size', help='How many task to ask for in a batch',
                        default=1, type=int)
    parser.add_argument('--fake-results', action='store_true',
                        help='Always return the same result regardless of the parameters'
                             '\n useful to debug and produce large amount of data quickly')

    arguments = sys.argv[1:]
    try:
        index = arguments.index('--')
        arguments = arguments[index + 1:]
    except ValueError:
        pass
    args = parser.parse_args(arguments)
    print(args)

    context = zmq.Context()
    print(f"Connecting to server ({args.master_address})...")
    socket = context.socket(zmq.REQ)
    socket.connect("tcp://" + args.master_address)

    WORKER_ID = str(uuid4())
    LAST_RESULT = []  # This is used to store the first render when --fake-result is set
    # while True:
    infos = query(socket, 'info', WORKER_ID)
    render_args = infos['render_args']

    rendering_class: Type[BaseRenderer] = getattr(importlib.import_module(render_args['engine']), 'Renderer')
    rendering_engine: BaseRenderer = rendering_class(args.root_folder, {**render_args, **vars(args)})

    evaluation_args = infos['evaluation_args']
    evaluator_class = getattr(importlib.import_module(evaluation_args['module']), 'Evaluator')
    evaluator: BaseEvaluator = evaluator_class(**infos['evaluation_args']['args'])

    # Gather all experiment-wide parameters
    inference_args = infos['inference']
    controls_args = infos['controls_args']
    inference_model = load_inference_model(inference_args)

    last_env = None
    last_model = None

    pbar = tqdm(smoothing=0)

    image_shapes = rendering_engine.declare_outputs()
    assert set(image_shapes.keys()).issubset(rendering_class.KEYS), \
        'Return value of declare_outputs() should match the declared KEYS var'
    eval_shapes = evaluator.declare_outputs()
    assert set(eval_shapes.keys()).issubset(evaluator_class.KEYS), \
        'Return value of declare_outputs() should match the declared KEYS var'
    declared_outputs = {
        **image_shapes,
        **eval_shapes,
        'output': (inference_args['output_shape'], 'float32'),
    }
    query(socket, 'decl', WORKER_ID, declared_outputs=declared_outputs)

    # These will be assigned in the if statement below
    model_uid = ''
    while True:
        job_description = query(socket, 'pull', WORKER_ID,
                                batch_size=args.batch_size,
                                last_environment=last_env,
                                last_model=last_model)
        parameters = job_description['params_to_render']

        if len(parameters) == 0:
            print("Nothing to do!", 'sleeping')
            time.sleep(1)
            continue

        for job in parameters:
            if LAST_RESULT:
                data = LAST_RESULT[0]
            else:
                current_env = job.environment
                current_model = job.model

                # We reload model and env if we got assigned to something
                # different this time
                if current_env != last_env or current_model != last_model:
                    print("==> [Loading new environment/model pair]")
                    loaded_env = rendering_engine.load_env(current_env)
                    loaded_model = rendering_engine.load_model(current_model)
                    model_uid = rendering_engine.get_model_uid(loaded_model)
                    rendering_engine.setup_render(loaded_model, loaded_env)
                    last_env = current_env
                    last_model = current_model

                controls_applier = ControlsApplier(job.control_order,
                                                job.render_args,
                                                controls_args,
                                                args.root_folder)

                scalar_label = evaluator.get_segmentation_label(model_uid)
                render_context = rendering_engine.get_context_dict(model_uid, scalar_label)
                controls_applier.apply_pre_controls(render_context)
                result = rendering_engine.render(model_uid,
                                                 loaded_model,
                                                 loaded_env)
                result['rgb'] = controls_applier.apply_post_controls(result['rgb'])
                controls_applier.unapply(render_context)

                with ch.no_grad():
                    result['rgb'] = result['rgb'][:3]
                    prediction, input_shape = inference_model(result['rgb'])

                lab = evaluator.get_target(model_uid, result)
                evaluation = evaluator.summary_stats(prediction, lab)
                assert evaluation.keys() == eval_shapes.keys(), \
                    'Outputs do not match declared outputs' \
                   f'{list(evaluation.keys())}, {list(eval_shapes.keys())}'
                out_shape =  inference_args['output_shape']
                prediction_tens = evaluator.to_tensor(prediction, out_shape, input_shape)
                data = {
                    'output': prediction_tens,
                    **result,
                    **evaluation
                }

                if args.fake_results:
                    LAST_RESULT.append(data)

            result_dtypes = {k: v[1] for (k, v) in declared_outputs.items()}
            query(socket, 'push', WORKER_ID, result_data=data,
                  result_dtypes=result_dtypes, job=job.id)
            pbar.update(1)



"""
Scheduling utils
================
"""

from threedb.utils import CyclicBuffer
import zmq
import numpy as np
import torch as ch
from typing import Dict, Any

def recv_array(socket, flags=0, copy=True, track=False):
    """recv a numpy array"""
    md = socket.recv_json(flags=flags)
    msg = socket.recv(flags=flags, copy=copy, track=track)
    buf = memoryview(msg)
    A = np.frombuffer(buf, dtype=md['dtype'])
    A = A.reshape(md['shape'])
    A = ch.from_numpy(A.copy())
    return A

def recv_into_buffer(socket: zmq.Socket, 
            cyclic_buffer: CyclicBuffer) -> Dict[str, Any]:
    main_message: Dict[str, Any] = socket.recv_json()

    if cyclic_buffer.initialized and ('result_keys' in main_message):
        result_keys = main_message['result_keys']
        buf_data = {}
        for result_key in result_keys:
            buf_data[result_key] = recv_array(socket)

        # outputs = recv_array(socket)
        # is_correct = socket.recv_pyobj()
        assert socket.recv_string() == 'done', 'Did not get done message'

        # ix = cyclic_buffer.allocate(images, outputs, is_correct)
        idx = cyclic_buffer.allocate(buf_data)
        main_message['result'] = idx

    return main_message


from setuptools import setup, find_packages
import distutils.cmd
import os
import distutils.log
import setuptools
import setuptools.command.build_py
import subprocess
from os import path
from subprocess import check_output

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='threedb-preview',
      version='0.0.7',
      description='A framework for analyzing computer vision models with simulated data ',
      url='https://github.com/3db/3db',
      author='3DB authors',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author_email='leclerc@mit.edu',
      license='MIT',
      install_requires=[
        'torch>=1.7.0',
        'torchvision',
        'cox',
        'robustness',
        'kornia',
        'scikit-image',
        'orjson',
        'opencv-python',
        'zmq',
        'tensorboard',
        'ipdb',
        'flatten-dict',
        'tqdm',
        'pytest',
        'pytest-xdist',
        'tensorboard',
        'typeguard',
        'pyyaml'],
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False)



"""
threedb.controls.blender.position
=================================

Control the object location. An example config file using this control can be found here:
`<https://github.com/3db/3db/tree/main/examples/unit_tests/position.yaml>`_.
"""

from typing import Any, Dict
from threedb.controls.base_control import PreProcessControl
from .utils import post_translate, cleanup_translate_containers
from ...try_bpy import mathutils

class PositionControl(PreProcessControl):
    """This control changes the position of the object (i.e. translates it)

    .. note::
        The change in position is relative to the object's original position.

    Continuous Parameters:

    - ``offset_x``: Translation compnonent along the world's x-axis (range: ``[-1, 1]``)
    - ``offset_y``: Translation compnonent along the world's y-axis (range: ``[-1, 1]``)
    - ``offset_z``: Translation compnonent along the world's z-axis (range: ``[-1, 1]``)

    .. admonition:: Example images

        .. thumbnail:: /_static/logs/position/images/image_1.png
            :width: 100
            :group: position

        .. thumbnail:: /_static/logs/position/images/image_2.png
            :width: 100
            :group: position

        .. thumbnail:: /_static/logs/position/images/image_3.png
            :width: 100
            :group: position

        .. thumbnail:: /_static/logs/position/images/image_4.png
            :width: 100
            :group: position

        .. thumbnail:: /_static/logs/position/images/image_5.png
            :width: 100
            :group: position

        Varying the position across the relevant parameters.
    """
    def __init__(self, root_folder: str):
        continuous_dims = {
            'offset_x': (-1., 1.),
            'offset_y': (-1., 1.),
            'offset_z': (-1., 1.),
        }
        super().__init__(root_folder, continuous_dims=continuous_dims)

    def apply(self, context: Dict[str, Any], control_args: Dict[str, Any]) -> None:
        no_err, msg = self.check_arguments(control_args)
        assert no_err, msg

        obj = context['object']
        post_translate(obj, mathutils.Vector([control_args['offset_x'],
                                   control_args['offset_y'],
                                   control_args['offset_z']]))

    def unapply(self, context: Dict[str, Any]) -> None:
        cleanup_translate_containers(context['object'])

Control = PositionControl



"""
threedb.scheduling.policy_controller
====================================
"""

from multiprocessing import Process, Queue
from queue import Empty
from collections import namedtuple
from uuid import uuid4
import numpy as np
from typing import List, Dict, Optional, Any
from threedb.scheduling.search_space import SearchSpace
from threedb.result_logging.logger_manager import LoggerManager
from threedb.utils import init_policy, CyclicBuffer


JobDescriptor = namedtuple("JobDescriptor", ['order', 'id', 'environment',
                                             'model', 'render_args',
                                             'control_order'])

class PolicyController(Process):

    def __init__(self, search_space: SearchSpace,
                       env_file: str,
                       model_name: str,
                       policy_args: Dict[str, Any],
                       logger_manager: LoggerManager,
                       result_buffer: CyclicBuffer):
        super().__init__()
        self.work_queue = Queue()
        self.result_queue = Queue()
        self.env_file = env_file
        self.model_name = model_name
        self.policy_args = policy_args
        self.search_space = search_space
        self.logger_manager = logger_manager
        self.result_buffer = result_buffer

    def pull_work(self):
        try:
            return self.work_queue.get(block=False)
        except Empty:
            return None

    def push_result(self, descriptor, result):
        self.result_queue.put((descriptor, result))

    def run(self):
        def render(args):
            # Posting the jobs to the queue

            all_descriptors = {}
            for i, (continuous_args, discrete_args) in enumerate(args):

                argument_dict, ctrl_list = self.search_space.unpack(continuous_args,
                                                                    discrete_args)
                current_id = str(uuid4())
                descriptor = JobDescriptor(order=i, id=current_id,
                                           render_args=argument_dict,
                                           control_order=ctrl_list,
                                           environment=self.env_file,
                                           model=self.model_name)
                all_descriptors[current_id] = descriptor
                self.work_queue.put(descriptor, block=True)

            client_results: List[Optional[dict]] = [None] * len(args)

            # Waiting and reordering the results
            for _ in range(len(args)):
                job_id, result_ix = self.result_queue.get(block=True)
                c_result = self.result_buffer[result_ix]
                descriptor = all_descriptors[job_id]

                self.logger_manager.log({
                    **descriptor._asdict(),
                    'result_ix': result_ix
                })
                client_results[descriptor.order] = {k: v.clone() for (k, v) in c_result.items()}
                self.result_buffer.free(result_ix, 1)

            result_keys = client_results[0].keys()
            stacked_results = {k: np.stack([res[k] for res in client_results]) for k in result_keys}
            return stacked_results

        policy = init_policy(self.policy_args)
        policy.run(render)



"""
threedb.controls.blender.imagenet_c
====================================

Apply corruptions to renderings. An example config file using this control can be found here:
`<https://github.com/3db/3db/tree/main/examples/unit_tests/imagenet_c.yaml>`_.
"""

from typing import Any, Dict
import torch as ch
from imagenet_c import corrupt
from ..base_control import PostProcessControl

class ImagenetCControl(PostProcessControl):
    """
    Applies the ImageNet-C corruptions from `<https://github.com/hendrycks/robustness>`_.

    Discrete Dimensions:

    - ``corruption_name``: The name of corruption that will be applied. Includes
        'gaussian_noise', 'shot_noise', 'impulse_noise', 'defocus_blur', 'glass_blur',
        'motion_blur', 'zoom_blur', 'snow', 'frost', 'fog', 'speckle_noise', 'gaussian_blur', 'spatter',
        'saturate', 'brightness', 'contrast', 'elastic_transform', 'pixelate', 'jpeg_compression'.
    - ``severity``: Imagenet-C severity parameter. (range: ``{0, 1, 2, 3, 4, 5}``)

    .. admonition:: Example images

        .. thumbnail:: /_static/logs/imagenet_c/images/image_1.png
            :width: 100
            :group: imagenet_c

        .. thumbnail:: /_static/logs/imagenet_c/images/image_2.png
            :width: 100
            :group: imagenet_c

        .. thumbnail:: /_static/logs/imagenet_c/images/image_3.png
            :width: 100
            :group: imagenet_c

        .. thumbnail:: /_static/logs/imagenet_c/images/image_4.png
            :width: 100
            :group: imagenet_c

        .. thumbnail:: /_static/logs/imagenet_c/images/image_5.png
            :width: 100
            :group: imagenet_c
        
        Examples of impulse noise application at various severity levels.
    """
    def __init__(self, root_folder: str):
        discrete_dims = {
            'severity': [1, 2, 3, 4, 5],
            'corruption_name': ['gaussian_noise', 'shot_noise', 'impulse_noise',
                                'defocus_blur', 'glass_blur', 'motion_blur',
                                'zoom_blur', 'snow', 'frost', 'fog',
                                'speckle_noise', 'gaussian_blur', 'spatter',
                                'saturate', 'brightness', 'contrast',
                                'elastic_transform', 'pixelate',
                                'jpeg_compression']
        }
        super().__init__(root_folder,
                         discrete_dims=discrete_dims)

    def apply(self, render: ch.Tensor, control_args: Dict[str, Any]) -> ch.Tensor:
        no_err, msg = self.check_arguments(control_args)
        assert no_err, msg

        sev, c_name = control_args['severity'], control_args['corruption_name']
        img = render.numpy()
        img = img.transpose(1, 2, 0)
        img = (img * 255).astype('uint8')
        img = corrupt(img, severity=sev, corruption_name=c_name)
        img = img.transpose(2, 0, 1)
        img = img.astype('float32') / 255
        return ch.from_numpy(img)


    def unapply(self, context: Dict[str, Any]) -> None:
        pass

Control = ImagenetCControl



from types import SimpleNamespace
try:
    import bpy, mathutils
except:
    # For type annotations
    bpy = SimpleNamespace(types=SimpleNamespace(
                            Scene=None,
                            Object=None,
                            Mesh=None))

    mathutils = SimpleNamespace(Vector=SimpleNamespace(
                            Euler=None))                            


"""
threedb.rendering.render_blender
================================

Implements the `Blender` rendering class that subclasses :class:`threedb.rendering.base_renderer.BaseRenderer`:.
This includes all the Blender-specific rendering settings, functions, and configs.
"""

import re
from glob import glob
from multiprocessing import cpu_count
from os import path
from tempfile import TemporaryDirectory
from typing import Tuple, Dict, Optional, Iterable, List, Any

from ..try_bpy import bpy

import cv2
import numpy as np
import torch as ch
from .base_renderer import BaseRenderer, RenderEnv, RenderObject
from .utils import ControlsApplier

IMAGE_FORMAT = 'png'

ENV_EXTENSIONS = ['blend', 'exr', 'hdr']

"""
Utility functions
"""
def _get_model_path(root_folder: str, model: str) -> str:
    return path.join(root_folder, 'blender_models', model)

def _get_env_path(root_folder: str, env: str) -> str:
    return path.join(root_folder, 'blender_environments', env)

def _setup_nice_PNG(input_node: Any) -> None:
    input_node.use_node_format = False
    input_node.format.file_format = "PNG"
    input_node.format.compression = 0
    input_node.format.color_depth = "16"

class Blender(BaseRenderer):
    NAME: str = 'Blender'
    KEYS: List[str] = ['rgb', 'segmentation', 'uv', 'depth']

    def __init__(self, root_dir: str, render_settings: Dict[str, Any], _ = None) -> None:
        super().__init__(root_dir, render_settings, ENV_EXTENSIONS)
        self.main_nodes = []
        self.post_process_nodes = []

    @staticmethod
    def enumerate_models(search_dir: str) -> List[str]:
        """
        Given a root folder, returns all .blend files in root/blender_models/
        """
        return [path.basename(x) for x in glob(_get_model_path(search_dir, '*.blend'))]

    @staticmethod
    def enumerate_environments(search_dir: str) -> List[str]:
        all_files = map(lambda x: path.basename(x), glob(_get_env_path(search_dir, '*.*')))
        return list(filter(lambda x: x.split('.')[-1] in ENV_EXTENSIONS, all_files))

    def declare_outputs(self) -> Dict[str, Tuple[List[int], str]]:
        imsize = [self.args['resolution'], self.args['resolution']]
        output_channels: Dict[str, Tuple[List[int], str]] = {'rgb': ([3, *imsize], 'float32')}
        if self.args['with_uv']:
            output_channels['uv'] = ([4, *imsize], 'float32')
        if self.args['with_depth']:
            output_channels['depth'] = ([4, *imsize], 'float32')
        if self.args['with_segmentation']:
            output_channels['segmentation'] = ([1, *imsize], 'int32')

        return output_channels

    def load_model(self, model: str) -> RenderObject:
        basename, filename = path.split(_get_model_path(self.root_dir, model))
        uid = filename.replace('.blend', '')
        blendfile = path.join(basename, uid + '.blend')
        section = "\\Object\\"
        object = uid

        filepath = uid + '.blend'
        directory = blendfile + section
        filename = object

        bpy.ops.wm.append(
            filepath=filepath,
            filename=filename,
            directory=directory)

        return bpy.data.objects[uid]

    def get_model_uid(self, loaded_model):
        return loaded_model.name

    def load_env(self, env: str) -> Optional[RenderEnv]:
        full_env_path = _get_env_path(self.root_dir, env)

        if env.endswith('.blend'): # full blender file
            bpy.ops.wm.open_mainfile(filepath=full_env_path)
        else:  # HDRI env
            bpy.ops.wm.read_homefile()
            bpy.data.objects.remove(bpy.data.objects["Cube"], do_unlink=True)
            bpy.data.objects.remove(bpy.data.objects["Light"], do_unlink=True)
            bpy.context.scene.render.film_transparent = False
            world = bpy.context.scene.world
            node_tree = world.node_tree
            output_node = world.node_tree.get_output_node('CYCLES')

            [node_tree.links.remove(x) for x in output_node.inputs[0].links]

            background_node = node_tree.nodes.new(type="ShaderNodeBackground")
            node_tree.links.new(background_node.outputs[0], output_node.inputs[0])

            img = bpy.data.images.load(full_env_path)
            env_texture_node = node_tree.nodes.new(type="ShaderNodeTexEnvironment")
            env_texture_node.image = img

            node_tree.links.new(env_texture_node.outputs[0], background_node.inputs[0])

    def _before_render(self) -> None:
        """
        Private utility function to be called from render()
        """
        # COLOR settings for render
        bpy.context.scene.display_settings.display_device = 'None'
        bpy.context.scene.sequencer_colorspace_settings.name = 'Raw'
        bpy.context.view_layer.update()
        bpy.context.scene.view_settings.view_transform = 'Standard'
        bpy.context.scene.view_settings.look = 'None'

        for node in self.main_nodes:
            node.mute = False

        for node in self.post_process_nodes:
            node.mute = True

    def _before_preprocess(self) -> None:
        # COLOR SETTINGS for RGB output
        bpy.context.scene.display_settings.display_device = 'sRGB'
        bpy.context.scene.sequencer_colorspace_settings.name = 'sRGB'
        bpy.context.view_layer.update()
        bpy.context.scene.view_settings.view_transform = 'Filmic'
        bpy.context.scene.view_settings.look = 'None'

        for node in self.main_nodes:
            node.mute = True

        for node in self.post_process_nodes:
            node.mute = False
    
    def _setup_render_device(self, scene: Any, prefs: Any):
        gpu_id: int = self.args['gpu_id']
        cpu_cores: Optional[int] = self.args['cpu_cores']

        cprefs = prefs.addons['cycles'].preferences
        cprefs.get_devices()  # important to update the list of devices

        for device in cprefs.devices:
            device.use = False

        if cpu_cores:
            scene.render.threads_mode = 'FIXED'
            cores_available = cpu_count()
            assert cpu_cores <= cores_available, f'Your machine has only {cpu_cores} cores.'
            scene.render.threads = max(1, cpu_cores)

        if gpu_id == -1:
            scene.cycles.device = 'CPU'
            cpu_devices = [x for x in cprefs.devices if x.type == 'CPU']
            cpu_devices[0].use = True
        else:
            scene.cycles.device = 'GPU'
            gpu_devices = [x for x in cprefs.devices if x.type == 'CUDA']
            if len(gpu_devices) != 0:
                gpu_devices[gpu_id].use = True
            else:
                raise ValueError('No GPUs found.')
        for d in cprefs.devices:
            print(f'Device {d.name} ({d.type}) used? {d.use}')

    def setup_render(self, model: Optional[RenderObject], env: Optional[RenderEnv]) -> None:
        while self.main_nodes:
            self.main_nodes.pop()
        while self.post_process_nodes:
            self.post_process_nodes.pop()

        scene = bpy.context.scene
        bpy.context.scene.render.engine = 'CYCLES'
        prefs = bpy.context.preferences

        self._setup_render_device(scene, prefs)

        bpy.context.scene.cycles.samples = self.args['samples']
        bpy.context.scene.render.tile_x = self.args['tile_size']
        bpy.context.scene.render.tile_y = self.args['tile_size']
        bpy.context.scene.render.resolution_x = self.args['resolution']
        bpy.context.scene.render.resolution_y = self.args['resolution']
        bpy.context.scene.render.use_persistent_data = True

        scene.use_nodes = True

        nodes = scene.node_tree.nodes
        links = scene.node_tree.links

        scene.view_layers["View Layer"].use_pass_uv = self.args['with_uv']
        bpy.context.scene.view_layers["View Layer"].use_pass_z = self.args['with_depth']
        bpy.context.scene.view_layers["View Layer"].use_pass_object_index = self.args['with_segmentation']

        scene.use_nodes = True
        scene.name = 'main_scene'

        for node in list(nodes):
            nodes.remove(node)

        layers_node = nodes.new(type="CompositorNodeRLayers")
        self.main_nodes.append(layers_node)

        file_output_node = nodes.new(type="CompositorNodeOutputFile")
        file_output_node.name = 'exr_output'
        self.main_nodes.append(file_output_node)
        file_output_node.format.file_format = "OPEN_EXR"
        file_output_node.format.exr_codec = 'NONE'
        output_slots = file_output_node.file_slots
        output_slots.remove(file_output_node.inputs[0])
        output_slots.new("render_exr")
        links.new(layers_node.outputs[0], file_output_node.inputs[0])

        if self.args['with_depth']:
            output_slots.new("depth")
            _setup_nice_PNG(file_output_node.file_slots["depth"])
            math_node = nodes.new(type="CompositorNodeMath")
            self.main_nodes.append(math_node)
            links.new(layers_node.outputs["Depth"], math_node.inputs[0])
            math_node.operation = "DIVIDE"
            math_node.inputs[1].default_value = self.args['max_depth']
            links.new(math_node.outputs[0], file_output_node.inputs["depth"])

        if self.args['with_uv']:
            output_slots.new("uv")
            _setup_nice_PNG(file_output_node.file_slots["uv"])
            links.new(layers_node.outputs["UV"], file_output_node.inputs["uv"])

        if self.args['with_segmentation']:
            output_slots.new("segmentation")
            _setup_nice_PNG(file_output_node.file_slots["segmentation"])
            file_output_node.file_slots["segmentation"].format.color_mode = "BW"
            math_node = nodes.new(type="CompositorNodeMath")
            self.main_nodes.append(math_node)
            links.new(layers_node.outputs["IndexOB"], math_node.inputs[0])
            math_node.operation = "DIVIDE"
            math_node.inputs[1].default_value = 65535.0
            links.new(math_node.outputs[0], file_output_node.inputs["segmentation"])

        input_image = nodes.new(type="CompositorNodeImage")
        self.post_process_nodes.append(input_image)
        input_image.name = "input_image"
        file_output_node = nodes.new(type="CompositorNodeOutputFile")
        file_output_node.name = "rgb_output"
        self.post_process_nodes.append(file_output_node)
        output_slots = file_output_node.file_slots
        output_slots.remove(file_output_node.inputs[0])
        output_slots.new("rgb")
        file_output_node.format.file_format = "PNG"
        file_output_node.format.compression = 0
        file_output_node.format.color_depth = "8"
        links.new(input_image.outputs["Image"], file_output_node.inputs["rgb"])
    
    def get_context_dict(self, model_uid: str, object_class: int) -> Dict[str, Any]:
        obj  = bpy.context.scene.objects[model_uid]

        # 0 is background so we shift everything by 1
        obj.pass_index = object_class + 1

        return {'object': obj}


    def render(self,
               model_uid: str,
               loaded_model: RenderObject, 
               loaded_env: RenderEnv) -> Dict[str, ch.Tensor]:
        output = {}

        with TemporaryDirectory() as temp_folder:
            scene = bpy.context.scene
            scene.node_tree.nodes['exr_output'].base_path = temp_folder
            self._before_render()
            bpy.ops.render.render(use_viewport=False, write_still=False)
            self._before_preprocess()
            written_file = glob(path.join(temp_folder, '*.exr'))
            blender_loaded_image = bpy.data.images.load(written_file[0])
            scene.node_tree.nodes["input_image"].image = blender_loaded_image
            scene.node_tree.nodes["rgb_output"].format.file_format = IMAGE_FORMAT.upper()
            scene.node_tree.nodes['rgb_output'].base_path = temp_folder
            bpy.ops.render.render(use_viewport=False, write_still=False)

            all_files = glob(path.join(temp_folder, "*.png"))

            for full_filename in all_files:
                name = re.sub(r'[0-9]+.png', '', path.basename(full_filename))
                img = cv2.imread(full_filename, cv2.IMREAD_UNCHANGED)

                if name == 'segmentation': 
                    img = img[:, :, None]  # Add extra dimension for the channel
                    img = img.astype('int32') - 1  # Go back from the 1 index to the 0 index
                    # We needed 1 index for the classes because we can only read images with
                    # positive integers
                elif img.dtype is np.dtype(np.uint16):
                    img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)
                    img = img.astype('float32') / (2**16 - 1)
                else:
                    img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)
                    img = img.astype('float32') / (2**8 - 1)

                output[name] = ch.from_numpy(img).permute(2, 0, 1)

            # Avoid memory leak by keeping all EXR rendered so far in memory
            bpy.data.images.remove(blender_loaded_image)

        return output
    
Renderer = Blender